__author__ = "Simon Nilsson"

import glob
import math
import os
from collections import defaultdict
from copy import deepcopy

import numpy as np
import pandas as pd

from simba.feature_extractors.perimeter_jit import jitted_hull
from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_fn_ext, read_df, write_df


class ExtractFeaturesFrom9bps(ConfigReader, FeatureExtractionMixin):
    """
    Extracts hard-coded set of features from pose-estimation data from single animals with 9 tracked body-parts.
    Results are stored in the `project_folder/csv/features_extracted` directory of the SimBA project.

    :parameter str config_path: path to SimBA project config file in Configparser format

    .. note::
       `Feature extraction tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/tutorial.md#step-5-extract-features>`__.
       `Expected pose configuration <https://github.com/sgoldenlab/simba/blob/master/simba/pose_configurations/schematics/4.png>`_

       .. image:: _static/img/pose_configurations/4.png
          :width: 150
          :align: center


    Examples
    ----------
    >>> feature_extractor = ExtractFeaturesFrom9bps(config_path='MyProjectConfig')
    >>> feature_extractor.run()
    """

    def __init__(self, config_path: str):
        FeatureExtractionMixin.__init__(self, config_path=config_path)
        ConfigReader.__init__(self, config_path=config_path)
        self.in_headers = self.get_feature_extraction_headers(
            pose="1 animal 9 body-parts"
        )
        self.mouse_p_headers = [x for x in self.in_headers if x[-2:] == "_p"]
        self.mouse_headers = [x for x in self.in_headers if x[-2:] != "_p"]

    def run(self):
        for file_cnt, file_path in enumerate(self.outlier_corrected_paths):
            video_timer = SimbaTimer(start=True)
            roll_windows = []
            _, self.video_name, _ = get_fn_ext(file_path)
            print(f"Extracting features for video {self.video_name}...")
            video_settings, self.px_per_mm, fps = self.read_video_info(
                video_name=self.video_name
            )
            for window in self.roll_windows_values:
                roll_windows.append(int(fps / window))
            self.roll_windows_values = [int(x) for x in self.roll_windows_values]
            self.in_data = (
                read_df(file_path, self.file_type)
                .fillna(0)
                .apply(pd.to_numeric)
                .reset_index(drop=True)
            )
            self.in_data = self.insert_default_headers_for_feature_extraction(
                df=self.in_data,
                headers=self.in_headers,
                pose_config="1 animal 9 body-parts",
                filename=file_path,
            )
            self.out_data = deepcopy(self.in_data)
            self.in_data_shifted = (
                self.out_data.shift(periods=1).add_suffix("_shifted").fillna(0)
            )
            self.in_data = (
                pd.concat([self.in_data, self.in_data_shifted], axis=1, join="inner")
                .fillna(0)
                .reset_index(drop=True)
            )
            mouse_arr = np.reshape(
                self.out_data[self.mouse_headers].values,
                (len(self.out_data / 2), -1, 2),
            ).astype(np.float32)
            print("Calculating hull features...")
            self.out_data["Mouse_poly_area"] = (
                jitted_hull(points=mouse_arr, target="perimeter") / self.px_per_mm
            )

            print("Calculating Euclidean distances... ")
            self.out_data["Nose_to_tail"] = self.euclidean_distance(
                self.in_data["Mouse1_nose_x"].values,
                self.in_data["Mouse1_tail_x"].values,
                self.in_data["Mouse1_nose_y"].values,
                self.in_data["Mouse1_tail_y"].values,
                self.px_per_mm,
            )
            self.out_data["Distance_feet"] = self.euclidean_distance(
                self.in_data["Mouse1_left_foot_x"].values,
                self.in_data["Mouse1_right_foot_x"].values,
                self.in_data["Mouse1_left_foot_y"].values,
                self.in_data["Mouse1_right_foot_y"].values,
                self.px_per_mm,
            )
            self.out_data["Distance_hands"] = self.euclidean_distance(
                self.in_data["Mouse1_left_hand_x"].values,
                self.in_data["Mouse1_right_hand_x"].values,
                self.in_data["Mouse1_left_hand_y"].values,
                self.in_data["Mouse1_right_hand_y"].values,
                self.px_per_mm,
            )
            self.out_data["Distance_ears"] = self.euclidean_distance(
                self.in_data["Mouse1_left_ear_x"].values,
                self.in_data["Mouse1_right_ear_x"].values,
                self.in_data["Mouse1_left_ear_y"].values,
                self.in_data["Mouse1_right_ear_y"].values,
                self.px_per_mm,
            )

            self.out_data["Distance_unilateral_left_hands_feet"] = (
                self.euclidean_distance(
                    self.in_data["Mouse1_left_foot_x"].values,
                    self.in_data["Mouse1_left_hand_x"].values,
                    self.in_data["Mouse1_left_foot_y"].values,
                    self.in_data["Mouse1_left_hand_y"].values,
                    self.px_per_mm,
                )
            )
            self.out_data["Distance_unilateral_right_hands_feet"] = (
                self.euclidean_distance(
                    self.in_data["Mouse1_right_foot_x"].values,
                    self.in_data["Mouse1_right_hand_x"].values,
                    self.in_data["Mouse1_right_foot_y"].values,
                    self.in_data["Mouse1_right_hand_y"].values,
                    self.px_per_mm,
                )
            )
            self.out_data["Distance_bilateral_left_foot_right_hand"] = (
                self.euclidean_distance(
                    self.in_data["Mouse1_left_foot_x"].values,
                    self.in_data["Mouse1_right_hand_x"].values,
                    self.in_data["Mouse1_left_foot_y"].values,
                    self.in_data["Mouse1_right_hand_y"].values,
                    self.px_per_mm,
                )
            )
            self.out_data["Distance_bilateral_right_foot_left_hand"] = (
                self.euclidean_distance(
                    self.in_data["Mouse1_right_foot_x"].values,
                    self.in_data["Mouse1_left_hand_x"].values,
                    self.in_data["Mouse1_right_foot_y"].values,
                    self.in_data["Mouse1_left_hand_y"].values,
                    self.px_per_mm,
                )
            )
            self.out_data["Distance_back_tail"] = self.euclidean_distance(
                self.in_data["Mouse1_back_x"].values,
                self.in_data["Mouse1_tail_x"].values,
                self.in_data["Mouse1_back_y"].values,
                self.in_data["Mouse1_tail_y"].values,
                self.px_per_mm,
            )
            self.out_data["Distance_back_nose"] = self.euclidean_distance(
                self.in_data["Mouse1_back_x"].values,
                self.in_data["Mouse1_nose_x"].values,
                self.in_data["Mouse1_back_y"].values,
                self.in_data["Mouse1_nose_y"].values,
                self.px_per_mm,
            )

            self.out_data["Movement_nose"] = self.euclidean_distance(
                self.in_data["Mouse1_nose_x_shifted"].values,
                self.in_data["Mouse1_nose_x"].values,
                self.in_data["Mouse1_nose_y_shifted"].values,
                self.in_data["Mouse1_nose_y"].values,
                self.px_per_mm,
            )
            self.out_data["Movement_back"] = self.euclidean_distance(
                self.in_data["Mouse1_back_x_shifted"].values,
                self.in_data["Mouse1_back_x"].values,
                self.in_data["Mouse1_back_y_shifted"].values,
                self.in_data["Mouse1_back_y"].values,
                self.px_per_mm,
            )
            self.out_data["Movement_left_ear"] = self.euclidean_distance(
                self.in_data["Mouse1_left_ear_x_shifted"].values,
                self.in_data["Mouse1_left_ear_x"].values,
                self.in_data["Mouse1_left_ear_y_shifted"].values,
                self.in_data["Mouse1_left_ear_y"].values,
                self.px_per_mm,
            )
            self.out_data["Movement_right_ear"] = self.euclidean_distance(
                self.in_data["Mouse1_right_ear_x_shifted"].values,
                self.in_data["Mouse1_right_ear_x"].values,
                self.in_data["Mouse1_right_ear_y_shifted"].values,
                self.in_data["Mouse1_right_ear_y"].values,
                self.px_per_mm,
            )
            self.out_data["Movement_left_foot"] = self.euclidean_distance(
                self.in_data["Mouse1_left_foot_x_shifted"].values,
                self.in_data["Mouse1_left_foot_x"].values,
                self.in_data["Mouse1_left_foot_y_shifted"].values,
                self.in_data["Mouse1_left_foot_y"].values,
                self.px_per_mm,
            )
            self.out_data["Movement_right_foot"] = self.euclidean_distance(
                self.in_data["Mouse1_right_foot_x_shifted"].values,
                self.in_data["Mouse1_right_foot_x"].values,
                self.in_data["Mouse1_right_foot_y_shifted"].values,
                self.in_data["Mouse1_right_foot_y"].values,
                self.px_per_mm,
            )
            self.out_data["Movement_tail"] = self.euclidean_distance(
                self.in_data["Mouse1_tail_x_shifted"].values,
                self.in_data["Mouse1_tail_x"].values,
                self.in_data["Mouse1_tail_y_shifted"].values,
                self.in_data["Mouse1_tail_y"].values,
                self.px_per_mm,
            )

            self.out_data["Movement_left_hand"] = self.euclidean_distance(
                self.in_data["Mouse1_left_hand_x_shifted"].values,
                self.in_data["Mouse1_left_hand_x"].values,
                self.in_data["Mouse1_left_hand_y_shifted"].values,
                self.in_data["Mouse1_left_hand_y"].values,
                self.px_per_mm,
            )
            self.out_data["Movement_right_hand"] = self.euclidean_distance(
                self.in_data["Mouse1_right_hand_x_shifted"].values,
                self.in_data["Mouse1_right_hand_x"].values,
                self.in_data["Mouse1_right_hand_y_shifted"].values,
                self.in_data["Mouse1_right_hand_y"].values,
                self.px_per_mm,
            )
            self.out_data["Mouse_polygon_size_change"] = (
                self.out_data["Mouse_poly_area"].shift(periods=1)
                - self.out_data["Mouse_poly_area"]
            )

            mouse_array = self.in_data[self.mouse_headers].to_numpy()
            self.hull_dict = defaultdict(list)
            for cnt, animal_frm in enumerate(mouse_array):
                animal_frm = np.reshape(animal_frm, (-1, 2))
                animal_dists = self.cdist(animal_frm, animal_frm)
                animal_dists = animal_dists[animal_dists != 0]
                self.hull_dict["Largest_euclidean_distance_hull"].append(
                    np.amax(animal_dists, initial=0) / self.px_per_mm
                )
                self.hull_dict["Smallest_euclidean_distance_hull"].append(
                    np.min(animal_dists, initial=0) / self.px_per_mm
                )
                self.hull_dict["Mean_euclidean_distance_hull"].append(
                    np.mean(animal_dists) / self.px_per_mm
                )
                self.hull_dict["Sum_euclidean_distance_hull"].append(
                    np.sum(animal_dists) / self.px_per_mm
                )
            for k, v in self.hull_dict.items():
                self.out_data[k] = v

            self.out_data["Total_movement_all_bodyparts"] = (
                self.out_data["Movement_nose"]
                + self.out_data["Movement_back"]
                + self.out_data["Movement_left_ear"]
                + self.out_data["Movement_right_ear"]
                + self.out_data["Movement_left_foot"]
                + self.out_data["Movement_right_foot"]
                + self.out_data["Movement_tail"]
                + self.out_data["Movement_left_hand"]
                + self.out_data["Movement_right_hand"]
            )

            print("Calculating rolling windows features...")
            for i in self.roll_windows_values:
                self.out_data[f"Nose_to_tail_median_{i}"] = (
                    self.out_data["Nose_to_tail"].rolling(i, min_periods=1).median()
                )
                self.out_data[f"Nose_to_tail_mean_{i}"] = (
                    self.out_data["Nose_to_tail"].rolling(i, min_periods=1).mean()
                )
                self.out_data[f"Nose_to_tail_sum_{i}"] = (
                    self.out_data["Nose_to_tail"].rolling(i, min_periods=1).sum()
                )

            for i in self.roll_windows_values:
                self.out_data[f"Distance_feet_median_{i}"] = (
                    self.out_data["Distance_feet"].rolling(i, min_periods=1).median()
                )
                self.out_data[f"Distance_feet_mean_{i}"] = (
                    self.out_data["Distance_feet"].rolling(i, min_periods=1).mean()
                )
                self.out_data[f"Distance_feet_sum_{i}"] = (
                    self.out_data["Distance_feet"].rolling(i, min_periods=1).sum()
                )

            for i in self.roll_windows_values:
                self.out_data[f"Distance_ears_median_{i}"] = (
                    self.out_data["Distance_ears"].rolling(i, min_periods=1).median()
                )
                self.out_data[f"Distance_ears_mean_{i}"] = (
                    self.out_data["Distance_ears"].rolling(i, min_periods=1).mean()
                )
                self.out_data[f"Distance_ears_sum_{i}"] = (
                    self.out_data["Distance_ears"].rolling(i, min_periods=1).sum()
                )

            for i in self.roll_windows_values:
                self.out_data[f"Distance_unilateral_left_hands_feet_median_{i}"] = (
                    self.out_data["Distance_unilateral_left_hands_feet"]
                    .rolling(i, min_periods=1)
                    .median()
                )
                self.out_data[f"Distance_unilateral_left_hands_feet_mean_{i}"] = (
                    self.out_data["Distance_unilateral_left_hands_feet"]
                    .rolling(i, min_periods=1)
                    .mean()
                )
                self.out_data[f"Distance_unilateral_left_hands_feet_sum_{i}"] = (
                    self.out_data["Distance_unilateral_left_hands_feet"]
                    .rolling(i, min_periods=1)
                    .sum()
                )

            for i in self.roll_windows_values:
                self.out_data[f"Distance_unilateral_right_hands_feet_median_{i}"] = (
                    self.out_data["Distance_unilateral_right_hands_feet"]
                    .rolling(i, min_periods=1)
                    .median()
                )
                self.out_data[f"Distance_unilateral_right_hands_feet_mean_{i}"] = (
                    self.out_data["Distance_unilateral_right_hands_feet"]
                    .rolling(i, min_periods=1)
                    .mean()
                )
                self.out_data[f"Distance_unilateral_right_hands_feet_sum_{i}"] = (
                    self.out_data["Distance_unilateral_right_hands_feet"]
                    .rolling(i, min_periods=1)
                    .sum()
                )

            for i in self.roll_windows_values:
                self.out_data[f"Distance_bilateral_left_foot_right_hand_median_{i}"] = (
                    self.out_data["Distance_bilateral_left_foot_right_hand"]
                    .rolling(i, min_periods=1)
                    .median()
                )
                self.out_data[f"Distance_bilateral_left_foot_right_hand_mean_{i}"] = (
                    self.out_data["Distance_bilateral_left_foot_right_hand"]
                    .rolling(i, min_periods=1)
                    .mean()
                )
                self.out_data[f"Distance_bilateral_left_foot_right_hand_sum_{i}"] = (
                    self.out_data["Distance_bilateral_left_foot_right_hand"]
                    .rolling(i, min_periods=1)
                    .sum()
                )

            for i in self.roll_windows_values:
                self.out_data[f"Distance_bilateral_right_foot_left_hand_median_{i}"] = (
                    self.out_data["Distance_bilateral_right_foot_left_hand"]
                    .rolling(i, min_periods=1)
                    .median()
                )
                self.out_data[f"Distance_bilateral_right_foot_left_hand_mean_{i}"] = (
                    self.out_data["Distance_bilateral_right_foot_left_hand"]
                    .rolling(i, min_periods=1)
                    .mean()
                )
                self.out_data[f"Distance_bilateral_right_foot_left_hand_sum_{i}"] = (
                    self.out_data["Distance_bilateral_right_foot_left_hand"]
                    .rolling(i, min_periods=1)
                    .sum()
                )

            for i in self.roll_windows_values:
                self.out_data[f"Distance_back_tail_median_{i}"] = (
                    self.out_data["Distance_back_tail"]
                    .rolling(i, min_periods=1)
                    .median()
                )
                self.out_data[f"Distance_back_tail_mean_{i}"] = (
                    self.out_data["Distance_back_tail"].rolling(i, min_periods=1).mean()
                )
                self.out_data[f"Distance_back_tail_sum_{i}"] = (
                    self.out_data["Distance_back_tail"].rolling(i, min_periods=1).sum()
                )

            for i in self.roll_windows_values:
                self.out_data[f"Distance_back_nose_median_{i}"] = (
                    self.out_data["Distance_back_nose"]
                    .rolling(i, min_periods=1)
                    .median()
                )
                self.out_data[f"Distance_back_nose_mean_{i}"] = (
                    self.out_data["Distance_back_nose"].rolling(i, min_periods=1).mean()
                )
                self.out_data[f"Distance_back_nose_sum_{i}"] = (
                    self.out_data["Distance_back_nose"].rolling(i, min_periods=1).sum()
                )

            for i in self.roll_windows_values:
                self.out_data[f"Movement_nose_median_{i}"] = (
                    self.out_data["Movement_nose"].rolling(i, min_periods=1).median()
                )
                self.out_data[f"Movement_nose_mean_{i}"] = (
                    self.out_data["Movement_nose"].rolling(i, min_periods=1).mean()
                )
                self.out_data[f"Movement_nose_sum_{i}"] = (
                    self.out_data["Movement_nose"].rolling(i, min_periods=1).sum()
                )

            for i in self.roll_windows_values:
                self.out_data[f"Movement_back_median_{i}"] = (
                    self.out_data["Movement_back"].rolling(i, min_periods=1).median()
                )
                self.out_data[f"Movement_back_mean_{i}"] = (
                    self.out_data["Movement_back"].rolling(i, min_periods=1).mean()
                )
                self.out_data[f"Movement_back_sum_{i}"] = (
                    self.out_data["Movement_back"].rolling(i, min_periods=1).sum()
                )

            for i in self.roll_windows_values:
                self.out_data[f"Movement_left_ear_median_{i}"] = (
                    self.out_data["Movement_left_ear"]
                    .rolling(i, min_periods=1)
                    .median()
                )
                self.out_data[f"Movement_left_ear_mean_{i}"] = (
                    self.out_data["Movement_left_ear"].rolling(i, min_periods=1).mean()
                )
                self.out_data[f"Movement_left_ear_sum_{i}"] = (
                    self.out_data["Movement_left_ear"].rolling(i, min_periods=1).sum()
                )

            for i in self.roll_windows_values:
                self.out_data[f"Movement_right_ear_median_{i}"] = (
                    self.out_data["Movement_right_ear"]
                    .rolling(i, min_periods=1)
                    .median()
                )
                self.out_data[f"Movement_right_ear_mean_{i}"] = (
                    self.out_data["Movement_right_ear"].rolling(i, min_periods=1).mean()
                )
                self.out_data[f"Movement_right_ear_sum_{i}"] = (
                    self.out_data["Movement_right_ear"].rolling(i, min_periods=1).sum()
                )

            for i in self.roll_windows_values:
                self.out_data[f"Movement_left_foot_median_{i}"] = (
                    self.out_data["Movement_left_foot"]
                    .rolling(i, min_periods=1)
                    .median()
                )
                self.out_data[f"Movement_left_foot_mean_{i}"] = (
                    self.out_data["Movement_left_foot"].rolling(i, min_periods=1).mean()
                )
                self.out_data[f"Movement_left_foot_sum_{i}"] = (
                    self.out_data["Movement_left_foot"].rolling(i, min_periods=1).sum()
                )

            for i in self.roll_windows_values:
                self.out_data[f"Movement_right_foot_median_{i}"] = (
                    self.out_data["Movement_right_foot"]
                    .rolling(i, min_periods=1)
                    .median()
                )
                self.out_data[f"Movement_right_foot_mean_{i}"] = (
                    self.out_data["Movement_right_foot"]
                    .rolling(i, min_periods=1)
                    .mean()
                )
                self.out_data[f"Movement_right_foot_sum_{i}"] = (
                    self.out_data["Movement_right_foot"].rolling(i, min_periods=1).sum()
                )

            for i in self.roll_windows_values:
                self.out_data[f"Movement_right_foot_median_{i}"] = (
                    self.out_data["Movement_right_foot"]
                    .rolling(i, min_periods=1)
                    .median()
                )
                self.out_data[f"Movement_right_foot_mean_{i}"] = (
                    self.out_data["Movement_right_foot"]
                    .rolling(i, min_periods=1)
                    .mean()
                )
                self.out_data[f"Movement_right_foot_sum_{i}"] = (
                    self.out_data["Movement_right_foot"].rolling(i, min_periods=1).sum()
                )

            for i in self.roll_windows_values:
                self.out_data[f"Movement_tail_median_{i}"] = (
                    self.out_data["Movement_tail"].rolling(i, min_periods=1).median()
                )
                self.out_data[f"Movement_tail_mean_{i}"] = (
                    self.out_data["Movement_tail"].rolling(i, min_periods=1).mean()
                )
                self.out_data[f"Movement_tail_sum_{i}"] = (
                    self.out_data["Movement_tail"].rolling(i, min_periods=1).sum()
                )

            for i in self.roll_windows_values:
                self.out_data[f"Movement_left_hand_median_{i}"] = (
                    self.out_data["Movement_left_hand"]
                    .rolling(i, min_periods=1)
                    .median()
                )
                self.out_data[f"Movement_left_hand_mean_{i}"] = (
                    self.out_data["Movement_left_hand"].rolling(i, min_periods=1).mean()
                )
                self.out_data[f"Movement_left_hand_sum_{i}"] = (
                    self.out_data["Movement_left_hand"].rolling(i, min_periods=1).sum()
                )

            for i in self.roll_windows_values:
                self.out_data[f"Movement_right_hand_median_{i}"] = (
                    self.out_data["Movement_right_hand"]
                    .rolling(i, min_periods=1)
                    .median()
                )
                self.out_data[f"Movement_right_hand_mean_{i}"] = (
                    self.out_data["Movement_right_hand"]
                    .rolling(i, min_periods=1)
                    .mean()
                )
                self.out_data[f"Movement_right_hand_sum_{i}"] = (
                    self.out_data["Movement_right_hand"].rolling(i, min_periods=1).sum()
                )

            for i in self.roll_windows_values:
                self.out_data[f"Total_movement_all_bodyparts_median_{i}"] = (
                    self.out_data["Total_movement_all_bodyparts"]
                    .rolling(i, min_periods=1)
                    .median()
                )
                self.out_data[f"Total_movement_all_bodyparts_mean_{i}"] = (
                    self.out_data["Total_movement_all_bodyparts"]
                    .rolling(i, min_periods=1)
                    .mean()
                )
                self.out_data[f"Total_movement_all_bodyparts_sum_{i}"] = (
                    self.out_data["Total_movement_all_bodyparts"]
                    .rolling(i, min_periods=1)
                    .sum()
                )

            for i in self.roll_windows_values:
                self.out_data[f"Mean_euclidean_distance_hull_median_{i}"] = (
                    self.out_data["Mean_euclidean_distance_hull"]
                    .rolling(i, min_periods=1)
                    .median()
                )
                self.out_data[f"Mean_euclidean_distance_hull_mean_{i}"] = (
                    self.out_data["Mean_euclidean_distance_hull"]
                    .rolling(i, min_periods=1)
                    .mean()
                )
                self.out_data[f"Mean_euclidean_distance_hull_sum_{i}"] = (
                    self.out_data["Mean_euclidean_distance_hull"]
                    .rolling(i, min_periods=1)
                    .sum()
                )

            for i in self.roll_windows_values:
                self.out_data[f"Smallest_euclidean_distance_hull_median_{i}"] = (
                    self.out_data["Smallest_euclidean_distance_hull"]
                    .rolling(i, min_periods=1)
                    .median()
                )
                self.out_data[f"Smallest_euclidean_distance_hull_mean_{i}"] = (
                    self.out_data["Smallest_euclidean_distance_hull"]
                    .rolling(i, min_periods=1)
                    .mean()
                )
                self.out_data[f"Smallest_euclidean_distance_hull_sum_{i}"] = (
                    self.out_data["Smallest_euclidean_distance_hull"]
                    .rolling(i, min_periods=1)
                    .sum()
                )

            for i in self.roll_windows_values:
                self.out_data[f"Largest_euclidean_distance_hull_median_{i}"] = (
                    self.out_data["Largest_euclidean_distance_hull"]
                    .rolling(i, min_periods=1)
                    .median()
                )
                self.out_data[f"Largest_euclidean_distance_hull_mean_{i}"] = (
                    self.out_data["Largest_euclidean_distance_hull"]
                    .rolling(i, min_periods=1)
                    .mean()
                )
                self.out_data[f"Largest_euclidean_distance_hull_sum_{i}"] = (
                    self.out_data["Largest_euclidean_distance_hull"]
                    .rolling(i, min_periods=1)
                    .sum()
                )

            print("Calculating angles...")
            self.out_data["Mouse_angle"] = self.angle3pt_serialized(
                data=self.out_data[
                    [
                        "Mouse1_nose_x",
                        "Mouse1_nose_y",
                        "Mouse1_back_x",
                        "Mouse1_back_y",
                        "Mouse1_tail_x",
                        "Mouse1_tail_y",
                    ]
                ].values
            )

            ########### DEVIATIONS ###########################################
            print("Calculating deviations...")
            self.out_data["Total_movement_all_bodyparts_both_deviation"] = (
                self.out_data["Total_movement_all_bodyparts"].mean()
                - self.out_data["Total_movement_all_bodyparts"]
            )
            self.out_data["Smallest_euclid_distances_hull_deviation"] = (
                self.out_data["Smallest_euclidean_distance_hull"].mean()
                - self.out_data["Smallest_euclidean_distance_hull"]
            )
            self.out_data["Largest_euclid_distances_hull_deviation"] = (
                self.out_data["Largest_euclidean_distance_hull"].mean()
                - self.out_data["Largest_euclidean_distance_hull"]
            )
            self.out_data["Mean_euclid_distances_hull_deviation"] = (
                self.out_data["Mean_euclidean_distance_hull"].mean()
                - self.out_data["Mean_euclidean_distance_hull"]
            )
            self.out_data["Movement_deviation_back"] = (
                self.out_data["Movement_back"].mean() - self.out_data["Movement_back"]
            )
            self.out_data["Polygon_deviation"] = (
                self.out_data["Mouse_poly_area"].mean()
                - self.out_data["Mouse_poly_area"]
            )

            for i in self.roll_windows_values:
                self.out_data[
                    f"Smallest_euclidean_distance_hull_mean_{i}_deviation"
                ] = (
                    self.out_data[f"Smallest_euclidean_distance_hull_mean_{i}"].mean()
                    - self.out_data[f"Smallest_euclidean_distance_hull_mean_{i}"]
                )

            for i in self.roll_windows_values:
                self.out_data[f"Largest_euclidean_distance_hull_mean_{i}_deviation"] = (
                    self.out_data[f"Largest_euclidean_distance_hull_mean_{i}"].mean()
                    - self.out_data[f"Largest_euclidean_distance_hull_mean_{i}"]
                )

            for i in self.roll_windows_values:
                self.out_data[f"Mean_euclidean_distance_hull_mean_{i}_deviation"] = (
                    self.out_data[f"Mean_euclidean_distance_hull_mean_{i}"].mean()
                    - self.out_data[f"Mean_euclidean_distance_hull_mean_{i}"]
                )

            for i in self.roll_windows_values:
                self.out_data[f"Total_movement_all_bodyparts_mean_{i}_deviation"] = (
                    self.out_data[f"Total_movement_all_bodyparts_mean_{i}"].mean()
                    - self.out_data[f"Total_movement_all_bodyparts_mean_{i}"]
                )

            self.out_data["Movement_percentile_rank"] = self.out_data[
                "Movement_back"
            ].rank(pct=True)

            for i in self.roll_windows_values:
                self.out_data[
                    f"Mean_euclidean_distance_hull_mean_{i}_percentile_rank"
                ] = (
                    self.out_data[f"Mean_euclidean_distance_hull_mean_{i}"].mean()
                    - self.out_data[f"Mean_euclidean_distance_hull_mean_{i}"]
                )

            for i in self.roll_windows_values:
                self.out_data[
                    f"Smallest_euclidean_distance_hull_mean_{i}_percentile_rank"
                ] = (
                    self.out_data[f"Smallest_euclidean_distance_hull_mean_{i}"].mean()
                    - self.out_data[f"Smallest_euclidean_distance_hull_mean_{i}"]
                )

            for i in self.roll_windows_values:
                self.out_data[
                    f"Largest_euclidean_distance_hull_mean_{i}_percentile_rank"
                ] = (
                    self.out_data[f"Largest_euclidean_distance_hull_mean_{i}"].mean()
                    - self.out_data[f"Largest_euclidean_distance_hull_mean_{i}"]
                )

            for i in self.roll_windows_values:
                self.out_data[
                    f"Total_movement_all_bodyparts_mean_{i}_percentile_rank"
                ] = (
                    self.out_data[f"Total_movement_all_bodyparts_mean_{i}"].mean()
                    - self.out_data[f"Total_movement_all_bodyparts_mean_{i}"]
                )

            ########### CALCULATE STRAIGHTNESS OF POLYLINE PATH: tortuosity  ###########################################
            print("Calculating path tortuosities...")
            as_strided = np.lib.stride_tricks.as_strided
            win_size = 3
            centroidList_Mouse1_x = as_strided(
                self.out_data.Mouse1_nose_x,
                (len(self.out_data) - (win_size - 1), win_size),
                (self.out_data.Mouse1_nose_x.values.strides * 2),
            )
            centroidList_Mouse1_y = as_strided(
                self.out_data.Mouse1_nose_y,
                (len(self.out_data) - (win_size - 1), win_size),
                (self.out_data.Mouse1_nose_y.values.strides * 2),
            )

            for k in range(len(self.roll_windows_values)):
                start = 0
                end = start + int(self.roll_windows_values[k])
                tortuosity_M1 = []
                for y in range(len(self.out_data)):
                    tortuosity_List_M1 = []
                    CurrCentroidList_Mouse1_x = centroidList_Mouse1_x[start:end]
                    CurrCentroidList_Mouse1_y = centroidList_Mouse1_y[start:end]
                    for i in range(len(CurrCentroidList_Mouse1_x)):
                        currMovementAngle_mouse1 = self.angle3pt(
                            CurrCentroidList_Mouse1_x[i][0],
                            CurrCentroidList_Mouse1_y[i][0],
                            CurrCentroidList_Mouse1_x[i][1],
                            CurrCentroidList_Mouse1_y[i][1],
                            CurrCentroidList_Mouse1_x[i][2],
                            CurrCentroidList_Mouse1_y[i][2],
                        )
                        tortuosity_List_M1.append(currMovementAngle_mouse1)
                    tortuosity_M1.append(sum(tortuosity_List_M1) / (2 * math.pi))
                    start += 1
                    end += 1
                self.out_data[f"Tortuosity_Mouse1_{self.roll_windows_values[k]}"] = (
                    tortuosity_M1
                )

            ########### CALC THE NUMBER OF LOW PROBABILITY DETECTIONS & TOTAL PROBABILITY VALUE FOR ROW###########################################
            print("Calculating pose probability scores...")
            self.out_data["Sum_probabilities"] = (
                self.out_data["Mouse1_left_ear_p"]
                + self.out_data["Mouse1_right_ear_p"]
                + self.out_data["Mouse1_left_hand_p"]
                + self.out_data["Mouse1_right_hand_p"]
                + self.out_data["Mouse1_left_foot_p"]
                + self.out_data["Mouse1_tail_p"]
                + self.out_data["Mouse1_right_foot_p"]
                + self.out_data["Mouse1_back_p"]
                + self.out_data["Mouse1_nose_p"]
            )
            self.out_data["Sum_probabilities_deviation"] = (
                self.out_data["Sum_probabilities"].mean()
                - self.out_data["Sum_probabilities"]
            )
            self.out_data["Sum_probabilities_deviation_percentile_rank"] = (
                self.out_data["Sum_probabilities_deviation"].rank(pct=True)
            )
            self.out_data["Sum_probabilities_percentile_rank"] = self.out_data[
                "Sum_probabilities_deviation_percentile_rank"
            ].rank(pct=True)
            results = pd.DataFrame(
                self.count_values_in_range(
                    data=self.out_data.filter(self.mouse_p_headers).values,
                    ranges=np.array([[0.0, 0.1], [0.0, 0.5], [0.0, 0.75]]),
                ),
                columns=[
                    "Low_prob_detections_0.1",
                    "Low_prob_detections_0.5",
                    "Low_prob_detections_0.75",
                ],
            )
            self.out_data = pd.concat([self.out_data, results], axis=1)

            self.out_data = self.out_data.reset_index(drop=True).fillna(0)
            save_path = os.path.join(
                self.save_dir, self.video_name + "." + self.file_type
            )
            write_df(df=self.out_data, file_type=self.file_type, save_path=save_path)

            video_timer.stop_timer()

            print(
                f"Feature extraction complete for {self.video_name} ({(file_cnt + 1)}/{len(self.files_found)} (elapsed time: {video_timer.elapsed_time_str}s)..."
            )

        self.timer.stop_timer()
        stdout_success(
            msg="All features extracted. Results stored in project_folder/csv/features_extracted directory",
            elapsed_time=self.timer.elapsed_time_str,
        )


#
# test = ExtractFeaturesFrom9bps(config_path='/Users/simon/Desktop/envs/troubleshooting/Emergence/project_folder/project_config.ini')
# test.run()



from tkinter import *

from simba.data_processors.mutual_exclusivity_corrector import \
    MutualExclusivityCorrector
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, DropDownMenu,
                                        Entry_Box, Label, LabelFrame)
from simba.utils.checks import check_float
from simba.utils.enums import Formats, Keys, Links
from simba.utils.errors import (DuplicationError, InvalidInputError,
                                NoSpecifiedOutputError)


class MutualExclusivityPupUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        ConfigReader.__init__(self, config_path=config_path)
        PopUpMixin.__init__(self, title="MUTUAL EXCLUSIVITY", size=(1000, 400))

        self.rule_cnt_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="EXCLUSIVITY RULES #",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.PATH_PLOTS.value,
        )
        self.rule_cnt_dropdown = DropDownMenu(
            self.rule_cnt_frm,
            "# RULES:",
            list(range(1, 21)),
            "25",
            com=self.create_rules_frames,
        )
        self.rule_cnt_dropdown.setChoices(1)
        self.rule_cnt_frm.grid(row=0, column=0, sticky="NW")
        self.rule_cnt_dropdown.grid(row=0, column=0, sticky="NW")
        self.create_rules_frames(rules_cnt=1)
        self.create_run_frm(run_function=self.run)
        self.main_frm.mainloop()

    def create_rules_frames(self, rules_cnt: int):
        if hasattr(self, "rule_definitions_frame"):
            self.rule_definitions_frame.destroy()
        self.rule_definitions_frame = LabelFrame(
            self.main_frm,
            text="RULE DEFINITIONS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=5,
        )
        self.rule_definitions_frame.grid(row=1, column=0, sticky="NW")
        Label(self.rule_definitions_frame, text="RULE #").grid(
            row=0, column=0, sticky=NW
        )
        for cnt, clf_name in enumerate(self.clf_names):
            Label(self.rule_definitions_frame, text=clf_name).grid(
                row=0, column=cnt + 1, sticky=NW
            )
        Label(self.rule_definitions_frame, text="WINNER").grid(
            row=0, column=self.clf_cnt + 1, sticky=NW
        )
        Label(self.rule_definitions_frame, text="THRESHOLD").grid(
            row=0, column=self.clf_cnt + 2, sticky=NW
        )
        Label(self.rule_definitions_frame, text="HIGHEST PROBABILITY").grid(
            row=0, column=self.clf_cnt + 3, sticky=NW
        )
        Label(self.rule_definitions_frame, text="TIE BREAK").grid(
            row=0, column=self.clf_cnt + 4, sticky=NW
        )
        Label(self.rule_definitions_frame, text="SKIP ON EQUAL").grid(
            row=0, column=self.clf_cnt + 5, sticky=NW
        )

        self.rules_dict = {}
        for rule_cnt in range(1, rules_cnt + 1):
            self.rules_dict[rule_cnt] = {}
            self.rules_dict[rule_cnt]["subordinates"] = {}
            Label(
                self.rule_definitions_frame,
                text=str(rule_cnt),
                font=Formats.LABELFRAME_HEADER_FORMAT.value,
            ).grid(row=rule_cnt, column=0, sticky=NW)
            for clf_cnt, clf_name in enumerate(self.clf_names):
                self.rules_dict[rule_cnt]["subordinates"][clf_name] = {}
                self.rules_dict[rule_cnt]["subordinates"][clf_name][
                    "cb_var"
                ] = BooleanVar()
                self.rules_dict[rule_cnt]["subordinates"][clf_name]["cb"] = Checkbutton(
                    self.rule_definitions_frame,
                    variable=self.rules_dict[rule_cnt]["subordinates"][clf_name][
                        "cb_var"
                    ],
                )
                self.rules_dict[rule_cnt]["subordinates"][clf_name]["cb"].grid(
                    row=rule_cnt, column=clf_cnt + 1, sticky=NW
                )

        for rule_cnt in range(1, rules_cnt + 1):
            self.rules_dict[rule_cnt]["winner_dropdown"] = DropDownMenu(
                self.rule_definitions_frame, "", self.clf_names, "1"
            )
            self.rules_dict[rule_cnt]["winner_dropdown"].disable()
            self.rules_dict[rule_cnt]["winner_dropdown"].setChoices(self.clf_names[0])
            self.rules_dict[rule_cnt]["winner_dropdown"].grid(
                row=rule_cnt, column=self.clf_cnt + 1, sticky=NW
            )
            self.rules_dict[rule_cnt]["threshold_entry"] = Entry_Box(
                self.rule_definitions_frame, "", "1"
            )
            self.rules_dict[rule_cnt]["threshold_entry"].entry_set(0.00)
            self.rules_dict[rule_cnt]["threshold_entry"].grid(
                row=rule_cnt, column=self.clf_cnt + 2, sticky=NW
            )
            self.rules_dict[rule_cnt]["threshold_entry"].set_state("disable")

        for rule_cnt in range(1, rules_cnt + 1):
            self.rules_dict[rule_cnt]["highest_var"] = BooleanVar(value=True)
            self.rules_dict[rule_cnt]["highest_cb"] = Checkbutton(
                self.rule_definitions_frame,
                variable=self.rules_dict[rule_cnt]["highest_var"],
                command=lambda k=rule_cnt: self.change_threshold_status(k),
            )
            self.rules_dict[rule_cnt]["highest_cb"].grid(
                row=rule_cnt, column=self.clf_cnt + 3, sticky=NW
            )

        for rule_cnt in range(1, rules_cnt + 1):
            self.rules_dict[rule_cnt]["tie_breaker_dropdown"] = DropDownMenu(
                self.rule_definitions_frame, "", self.clf_names, "1"
            )
            self.rules_dict[rule_cnt]["tie_breaker_dropdown"].setChoices(
                self.clf_names[0]
            )
            self.rules_dict[rule_cnt]["tie_breaker_dropdown"].grid(
                row=rule_cnt, column=self.clf_cnt + 4, sticky=NW
            )

        for rule_cnt in range(1, rules_cnt + 1):
            self.rules_dict[rule_cnt]["skip_on_equal_var"] = BooleanVar(value=False)
            self.rules_dict[rule_cnt]["skip_on_equal_cb"] = Checkbutton(
                self.rule_definitions_frame,
                variable=self.rules_dict[rule_cnt]["skip_on_equal_var"],
                command=lambda k=rule_cnt: self.change_winner_status(k),
            )
            self.rules_dict[rule_cnt]["skip_on_equal_cb"].grid(
                row=rule_cnt, column=self.clf_cnt + 5, sticky=NW
            )

    def change_winner_status(self, rule_cnt):
        if self.rules_dict[rule_cnt]["skip_on_equal_var"].get():
            self.rules_dict[rule_cnt]["winner_dropdown"].disable()
        else:
            self.rules_dict[rule_cnt]["winner_dropdown"].enable()

    def change_threshold_status(self, rule_cnt):
        if self.rules_dict[rule_cnt]["highest_var"].get():
            self.rules_dict[rule_cnt]["winner_dropdown"].disable()
            self.rules_dict[rule_cnt]["threshold_entry"].set_state("disable")
            self.rules_dict[rule_cnt]["tie_breaker_dropdown"].enable()
            self.rules_dict[rule_cnt]["skip_on_equal_cb"].config(state=NORMAL)
        else:
            self.rules_dict[rule_cnt]["winner_dropdown"].enable()
            self.rules_dict[rule_cnt]["threshold_entry"].set_state("normal")
            self.rules_dict[rule_cnt]["tie_breaker_dropdown"].disable()
            self.rules_dict[rule_cnt]["skip_on_equal_cb"].config(state=DISABLED)

    def run(self):
        rules = {}
        for rule_cnt, rule_data in self.rules_dict.items():
            rules[rule_cnt] = {}

            if rule_data["highest_var"].get():
                rules[rule_cnt]["rule_type"] = "highest_probability"
                rules[rule_cnt]["skip_files_with_identical"] = rule_data[
                    "skip_on_equal_var"
                ].get()
                if not rules[rule_cnt]["skip_files_with_identical"]:
                    rules[rule_cnt]["tie_breaker"] = rule_data[
                        "tie_breaker_dropdown"
                    ].getChoices()
                else:
                    rules[rule_cnt]["tie_breaker"] = None
                rules[rule_cnt]["subordinates"] = []
                for subordinate_clf in rule_data["subordinates"].keys():
                    subordinate_rule = rule_data["subordinates"][subordinate_clf][
                        "cb_var"
                    ].get()
                    if subordinate_rule:
                        rules[rule_cnt]["subordinates"].append(subordinate_clf)
                if len(rules[rule_cnt]["subordinates"]) < 2:
                    raise InvalidInputError(
                        msg=f"Less than two classifiers ticked for rule #{rule_cnt}",
                        source=self.__class__.__name__,
                    )
                if (
                    rules[rule_cnt]["tie_breaker"]
                    not in rules[rule_cnt]["subordinates"]
                    and rules[rule_cnt]["tie_breaker"] is not None
                ):
                    raise InvalidInputError(
                        msg=f'For rule {rule_cnt}, the tie-breaker ({rules[rule_cnt]["tie_breaker"]}) is not one of the selected classifiers.',
                        source=self.__class__.__name__,
                    )

            else:
                rules[rule_cnt]["rule_type"] = "clf_threshold"
                check_float(
                    name=f"Rule {rule_cnt}",
                    value=rule_data["threshold_entry"].entry_get,
                    max_value=1.00,
                    min_value=0.00,
                    raise_error=True,
                )
                rules[rule_cnt]["threshold"] = float(
                    rule_data["threshold_entry"].entry_get
                )
                rules[rule_cnt]["winner"] = rule_data["winner_dropdown"].getChoices()
                rules[rule_cnt]["subordinates"] = []
                for subordinate_clf in rule_data["subordinates"].keys():
                    subordinate_rule = rule_data["subordinates"][subordinate_clf][
                        "cb_var"
                    ].get()
                    if subordinate_rule:
                        rules[rule_cnt]["subordinates"].append(subordinate_clf)
                if len(rules[rule_cnt]["subordinates"]) < 2:
                    raise InvalidInputError(
                        msg=f"Less than two classifiers ticked for rule #{rule_cnt}",
                        source=self.__class__.__name__,
                    )
                if rules[rule_cnt]["winner"] not in rules[rule_cnt]["subordinates"]:
                    raise InvalidInputError(
                        msg=f'For rule {rule_cnt}, the winner ({rules[rule_cnt]["winner"]}) is not one of the selected classifiers.',
                        source=self.__class__.__name__,
                    )
                rules[rule_cnt]["subordinates"] = [
                    x
                    for x in rules[rule_cnt]["subordinates"]
                    if x != rules[rule_cnt]["winner"]
                ]

        corrector = MutualExclusivityCorrector(
            rules=rules, config_path=self.config_path
        )
        corrector.run()


# test = MutualExclusivityPupUp(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')



import time
from functools import cmp_to_key

import numba
import numpy as np
from numba import njit, prange, typeof, types
from numba.core.errors import TypingError
from numba.extending import overload, register_jitable
from numba.typed import List
from scipy.spatial import ConvexHull


@overload(ConvexHull)
def impl_convex_hull(data):
    def func(data):
        def brute_hull(data):
            s = []
            mid = [0, 0]
            for i in range(data.shape[0]):
                for j in range(i + 1, data.shape[0]):
                    x1, x2 = data[i][0], data[j][0]
                    y1, y2 = data[i][1], data[j][1]
                    a1, b1, c1 = y1 - y2, x2 - x1, x1 * y2 - y1 * x2
                    pos, neg = 0, 0
                    for k in range(data.shape[0]):
                        if a1 * data[k][0] + b1 * data[k][1] + c1 <= 0:
                            neg += 1
                        if a1 * data[k][0] + b1 * data[k][1] + c1 >= 0:
                            pos += 1
                    if pos == data.shape[0] or neg == data.shape[0]:
                        s.append(list(data[i]))
                        s.append(list(data[j]))

            s_unique = []
            for i in s:
                if i not in s_unique:
                    s_unique.append(i)

            n = len(s_unique)
            for i in range(n):
                mid[0] += s_unique[i][0]
                mid[1] += s_unique[i][1]
                s_unique[i][0] *= n
                s_unique[i][1] *= n

            # s = np.unique(np.array(s), axis=0)

        def division(frm_data):
            if frm_data.shape[0] <= 5:
                brute_hull(frm_data)

        for frm_cnt in range(data.shape[0]):
            frm_data = data[frm_cnt, :, :]
            frm_data = frm_data[frm_data[:, 0].argsort()]
            _ = division(frm_data)

    return func


data = np.random.randint(1, 10, size=(10, 4, 2))


@njit
def test_find_left(data):
    return ConvexHull(data)


start = time.time()
test_find_left(data)
end = time.time()
print(end - start)



__author__ = "Simon Nilsson"

import functools
import multiprocessing
import os
import platform

import cv2

from simba.data_processors.directing_other_animals_calculator import \
    DirectingOtherAnimalsAnalyzer
from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import check_file_exist_and_readable
from simba.utils.data import create_color_palettes
from simba.utils.enums import Paths
from simba.utils.lookups import get_color_dict
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (concatenate_videos_in_folder, get_fn_ext,
                                    get_video_meta_data, read_df)
from simba.utils.warnings import NoDataFoundWarning


class DirectingOtherAnimalsVisualizerMultiprocess(ConfigReader, PlottingMixin):
    """
    Class for visualizing when animals are directing towards body-parts of other animals using multiprocessing.

    .. important::
       Requires the pose-estimation data for the left ear, right ears and nose of individual animals.

    .. note::
        `Example of expected output <https://www.youtube.com/watch?v=d6pAatreb1E&list=PLi5Vwf0hhy1R6NDQJ3U28MOUJPfl2YWYl&index=22`_.

        `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Scenario2.md#visualizing-data-tables`>_.

        .. image:: _static/img/directing_other_animals.png
           :width: 450
           :align: center


    :parameter str config_path: path to SimBA project config file in Configparser format
    :parameter str video_name: Video to visualize directionality for in the project_folder/videos directory (e.g., ``My_video.mp4``)
    :parameter dict style_attr: Video style attribitions (colors and sizes etc.)
    :parameter dict core_cnt: How many cores to use to create the video.

    Examples
    -----
    >>> style_attr = {'Show_pose': True, 'Pose_circle_size': 3, "Direction_color": 'Random', 'Direction_thickness': 4, 'Highlight_endpoints': True}
    >>> directing_visualizer = DirectingOtherAnimalsVisualizerMultiprocess(config_path='project_folder/project_config.ini', video_name='Testing_Video_3.mp4', style_attr=style_attr)
    >>> directing_visualizer.run()
    """

    def __init__(
        self, config_path: str, data_path: str, style_attr: dict, core_cnt: int
    ):
        ConfigReader.__init__(self, config_path=config_path)
        PlottingMixin.__init__(self)
        if platform.system() == "Darwin":
            multiprocessing.set_start_method("spawn", force=True)

        self.data_path = data_path
        _, self.video_name, _ = get_fn_ext(self.data_path)
        self.direction_analyzer = DirectingOtherAnimalsAnalyzer(
            config_path=config_path,
            bool_tables=False,
            summary_tables=False,
            aggregate_statistics_tables=False,
        )
        self.direction_analyzer.outlier_corrected_paths = [data_path]
        self.direction_analyzer.run()
        self.direction_analyzer._transpose_results_to_df()
        self.style_attr, self.pose_colors, self.core_cnt = style_attr, [], core_cnt
        self.colors = get_color_dict()
        if self.style_attr["Show_pose"]:
            self.pose_colors = create_color_palettes(
                self.animal_cnt, int(len(self.x_cols) + 1)
            )
        if self.style_attr["Direction_color"] == "Random":
            self.direction_colors = create_color_palettes(1, int(self.animal_cnt**2))
        else:
            self.direction_colors = [self.colors[self.style_attr["Direction_color"]]]
        self.data_dict = self.direction_analyzer.directionality_df_dict
        self.video_path = self.find_video_of_file(self.video_dir, self.video_name)
        self.save_directory = os.path.join(
            self.project_path, Paths.DIRECTING_BETWEEN_ANIMALS_OUTPUT_PATH.value
        )
        if not os.path.exists(self.save_directory):
            os.makedirs(self.save_directory)
        self.data_path = os.path.join(
            self.outlier_corrected_dir, self.video_name + "." + self.file_type
        )
        check_file_exist_and_readable(file_path=self.data_path)
        print(f"Processing video {self.video_name}...")

    def run(self):
        """
        Method to create directionality videos. Results are stored in
        the `project_folder/frames/output/ROI_directionality_visualize` directory of the SimBA project

        Returns
        ----------
        None
        """

        self.data_df = read_df(self.data_path, file_type=self.file_type).fillna(0)
        self.save_path = os.path.join(self.save_directory, self.video_name + ".mp4")
        self.save_temp_path = os.path.join(self.save_directory, "temp")
        if os.path.exists(self.save_temp_path):
            self.remove_a_folder(folder_dir=self.save_temp_path)
        os.makedirs(self.save_temp_path)
        self.cap = cv2.VideoCapture(self.video_path)
        self.video_meta_data = get_video_meta_data(self.video_path)
        self.video_data = self.data_dict[self.video_name]
        if self.video_name in list(self.video_data["Video"]):
            self.__create_video()
        else:
            NoDataFoundWarning(
                msg=f"SimBA skipping video {self.video_name}: No animals are directing each other in the video."
            )

    def __create_video(self):
        video_timer = SimbaTimer(start=True)
        data_arr, frm_per_core = self.split_and_group_df(
            df=self.data_df, splits=self.core_cnt, include_split_order=True
        )
        print(
            "Creating directing images, multiprocessing (determined chunksize: {}, cores: {})...".format(
                str(self.multiprocess_chunksize), str(self.core_cnt)
            )
        )
        with multiprocessing.Pool(
            self.core_cnt, maxtasksperchild=self.maxtasksperchild
        ) as pool:
            constants = functools.partial(
                self.directing_animals_mp,
                directionality_data=self.video_data,
                video_meta_data=self.video_meta_data,
                style_attr=self.style_attr,
                save_temp_dir=self.save_temp_path,
                video_path=self.video_path,
                bp_names=self.animal_bp_dict,
                colors=self.direction_colors,
            )
            for cnt, result in enumerate(
                pool.imap(constants, data_arr, chunksize=self.multiprocess_chunksize)
            ):
                print(
                    "Image {}/{}, Video {}...".format(
                        str(int(frm_per_core * (result + 1))),
                        str(len(self.data_df)),
                        self.video_name,
                    )
                )

            concatenate_videos_in_folder(
                in_folder=self.save_temp_path,
                save_path=self.save_path,
                video_format="mp4",
            )
            video_timer.stop_timer()
            pool.terminate()
            pool.join()
            self.timer.stop_timer()
            stdout_success(
                msg=f"Video {self.video_name} created. Video saved at {self.save_path}",
                elapsed_time=self.timer.elapsed_time_str,
            )


# style_attr = {'Show_pose': True,
#               'Pose_circle_size': 4,
#               "Direction_color": 'Random',
#               'Direction_thickness': 5,
#               'Highlight_endpoints': True,
#               'Polyfill': False}
# test = DirectingOtherAnimalsVisualizerMultiprocess(config_path='/Users/simon/Desktop/envs/troubleshooting/khan/project_folder/project_config.ini',
#                                                    data_path='/Users/simon/Desktop/envs/troubleshooting/khan/project_folder/csv/outlier_corrected_movement_location/stitched.csv',
#                                                    style_attr=style_attr,
#                                                    core_cnt=5)
#
# test.run()


# style_attr = {'Show_pose': True,
#               'Pose_circle_size': 3,
#               "Direction_color": 'Random',
#               'Direction_thickness': 4,
#               'Highlight_endpoints': True,
#               'Polyfill': True}
# test = DirectingOtherAnimalsVisualizerMultiprocess(config_path='/Users/simon/Desktop/envs/troubleshooting/sleap_5_animals/project_folder/project_config.ini',
#                                                    data_path='/Users/simon/Desktop/envs/troubleshooting/sleap_5_animals/project_folder/csv/outlier_corrected_movement_location/Testing_Video_3.csv',
#                                                    style_attr=style_attr,
#                                                    core_cnt=5)
#
# test.run()


# style_attr = {'Show_pose': True, 'Pose_circle_size': 3, "Direction_color": 'Random', 'Direction_thickness': 4, 'Highlight_endpoints': True, 'Polyfill': True}
# test = DirectingOtherAnimalsVisualizerMultiprocess(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                        data_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/outlier_corrected_movement_location/Together_1.csv',
#                                        style_attr=style_attr,
#                                                    core_cnt=5)
#
# test.run()



__author__ = "Simon Nilsson"

import os
import sys
from enum import Enum
from pathlib import Path

import cv2
import numpy as np
import pkg_resources

import simba


class ConfigKey(Enum):
    GENERAL_SETTINGS = "General settings"
    PROJECT_PATH = "project_path"
    SML_SETTINGS = "SML settings"
    VIDEO_INFO_CSV = "video_info.csv"
    FOLDER_PATH = "folder_path"
    FILE_TYPE = "workflow_file_type"
    TARGET_CNT = "No_targets"
    ANIMAL_CNT = "animal_no"
    PROJECT_NAME = "project_name"
    OS = "OS_system"
    MODEL_DIR = "model_dir"
    THRESHOLD_SETTINGS = "threshold_settings"
    MIN_BOUT_LENGTH = "Minimum_bout_lengths"
    FRAME_SETTINGS = "Frame settings"
    LINE_PLOT_SETTINGS = "Line plot settings"
    PATH_PLOT_SETTINGS = "Path plot settings"
    DISTANCE_PLOT_SETTINGS = "Distance plot"
    HEATMAP_SETTINGS = "Heatmap settings"
    ROI_SETTINGS = "ROI settings"
    DIRECTIONALITY_SETTINGS = "Directionality settings"
    BODYPART_DIRECTION_VALUE = "bodypart_direction"
    PROBABILITY_THRESHOLD = "probability_threshold"
    PROCESS_MOVEMENT_SETTINGS = "process movements"
    CREATE_ENSEMBLE_SETTINGS = "create ensemble settings"
    VALIDATION_SETTINGS = "validation/run model"
    MULTI_ANIMAL_ID_SETTING = "Multi animal IDs"
    MULTI_ANIMAL_IDS = "ID_list"
    OUTLIER_SETTINGS = "Outlier settings"
    POSE_SETTING = "pose_estimation_body_parts"
    RF_JOBS = "RF_n_jobs"
    VALIDATION_VIDEO = "generate_validation_video"
    MOVEMENT_CRITERION = "movement_criterion"
    LOCATION_CRITERION = "location_criterion"
    ROI_ANIMAL_CNT = "no_of_animals"
    DISTANCE_MM = "distance_mm"
    SKLEARN_BP_PROB_THRESH = "bp_threshold_sklearn"


class Paths(Enum):
    INPUT_CSV = Path("csv/input_csv/")
    LINE_PLOT_DIR = Path("frames/output/line_plot/")
    VIDEO_INFO = Path("logs/video_info.csv")
    OUTLIER_CORRECTED = Path("csv/outlier_corrected_movement_location/")
    OUTLIER_CORRECTED_MOVEMENT = Path("csv/outlier_corrected_movement/")
    MACHINE_RESULTS_DIR = Path("csv/machine_results/")
    SKLEARN_RESULTS = Path("frames/output/sklearn_results/")
    CLUSTER_EXAMPLES = Path("frames/output/cluster_examples/")
    CLF_VALIDATION_DIR = Path("frames/output/classifier_validation/")
    CLF_DATA_VALIDATION_DIR = Path("csv/validation/")
    TEST_PATH = "/Users/simon/Desktop/envs/simba_dev/simba/"
    SINGLE_CLF_VALIDATION = Path("frames/output/validation/")
    INPUT_FRAMES_DIR = Path("frames/input/")
    ICON_ASSETS = Path("assets/icons/")
    DATA_TABLE = Path("frames/output/live_data_table/")
    ROI_FEATURES = Path("frames/output/ROI_features/")
    ROI_ANALYSIS = Path("frames/output/ROI_analysis/")
    ANNOTATED_FRAMES_DIR = Path("frames/output/annotated_frames/")
    DIRECTIONALITY_DF_DIR = Path("logs/directionality_dataframes/")
    BODY_PART_DIRECTIONALITY_DF_DIR = Path("logs/body_part_directionality_dataframes/")
    DIRECTING_ANIMALS_OUTPUT_PATH = Path("frames/output/ROI_directionality_visualize/")
    DIRECTING_BETWEEN_ANIMALS_OUTPUT_PATH = Path("frames/output/Directing_animals/")
    DIRECTING_BETWEEN_ANIMAL_BODY_PART_OUTPUT_PATH = Path(
        "frames/output/Body_part_directing_animals/"
    )
    BP_NAMES = Path("logs/measures/pose_configs/bp_names/project_bp_names.csv")
    SIMBA_BP_CONFIG_PATH = Path("pose_configurations/bp_names/bp_names.csv")
    SIMBA_SHAP_CATEGORIES_PATH = Path(
        "assets/shap/feature_categories/shap_feature_categories.csv"
    )
    SIMBA_FEATURE_EXTRACTION_COL_NAMES_PATH = Path(
        "assets/lookups/feature_extraction_headers.csv"
    )
    SIMBA_NO_ANIMALS_PATH = Path("pose_configurations/no_animals/no_animals.csv")
    SIMBA_SHAP_IMG_PATH = Path("assets/shap/")
    SCHEMATICS = Path("pose_configurations/schematics/")
    PROJECT_POSE_CONFIG_NAMES = Path(
        "pose_configurations/configuration_names/pose_config_names.csv"
    )
    CONCAT_VIDEOS_DIR = Path("frames/output/merged/")
    GANTT_PLOT_DIR = Path("frames/output/gantt_plots/")
    HEATMAP_CLF_LOCATION_DIR = Path("frames/output/heatmaps_classifier_locations/")
    HEATMAP_LOCATION_DIR = Path("frames/output/heatmaps_locations/")
    FRAMES_OUTPUT_DIR = Path("frames/output/")
    FEATURES_EXTRACTED_DIR = Path("csv/features_extracted/")
    TARGETS_INSERTED_DIR = Path("csv/targets_inserted/")
    PATH_PLOT_DIR = Path("frames/output/path_plots")
    ABOUT_ME = Path("assets/img/about_me.png")
    PROBABILITY_PLOTS_DIR = Path("frames/output/probability_plots/")
    ROI_DEFINITIONS = Path("measures/ROI_definitions.h5")
    DETAILED_ROI_DATA_DIR = Path("logs/Detailed_ROI_data/")
    SHAP_LOGS = Path("logs/shap/")
    SPLASH_PATH_WINDOWS = Path("assets/img/splash.png")
    SPLASH_PATH_LINUX = Path("assets/img/splash.PNG")
    SPLASH_PATH_MOVIE = Path("assets/img/splash_2024.mp4")
    BG_IMG_PATH = Path("assets/img/bg_2024.png")
    LOGO_ICON_WINDOWS_PATH = Path("assets/icons/SimBA_logo.ico")
    LOGO_ICON_DARWIN_PATH = Path("assets/icons/SimBA_logo.png")
    UNSUPERVISED_MODEL_NAMES = Path("assets/lookups/model_names.parquet")
    CRITICAL_VALUES = Path("simba/assets/lookups/critical_values_05.pickle")


class Formats(Enum):
    MP4_CODEC = "mp4v"
    AVI_CODEC = "XVID"
    BATCH_CODEC = "libx264"
    LABELFRAME_HEADER_FORMAT = ("Helvetica", 12, "bold")
    LABELFRAME_HEADER_CLICKABLE_FORMAT = ("Helvetica", 12, "bold", "underline")
    LABELFRAME_HEADER_CLICKABLE_COLOR = f"#{5:02x}{99:02x}{193:02x}"
    CSV = "csv"
    PARQUET = "parquet"
    PICKLE = "pickle"
    PERIMETER = "perimeter"
    AREA = "area"
    H5 = "h5"
    ROOT_WINDOW_SIZE = "750x750"
    FONT = cv2.FONT_HERSHEY_TRIPLEX
    TKINTER_FONT = ("Rockwell", 11)
    DLC_NETWORK_FILE_NAMES = [
        "dlc_resnet50",
        "dlc_resnet_50",
        "dlc_dlcrnetms5",
        "dlc_effnet_b0",
        "dlc_resnet101",
    ]
    DLC_FILETYPES = {
        "skeleton": ["sk.h5", "sk_filtered.h5"],
        "box": ["bx.h5", "bx_filtered.h5"],
        "ellipse": ["el.h5", "el_filtered.h5"],
    }


class Options(Enum):
    ROLLING_WINDOW_DIVISORS = [2, 5, 6, 7.5, 15]
    CLF_MODELS = ["RF", "GBC", "XGBoost"]
    CLF_MAX_FEATURES = ["sqrt", "log", "None"]
    CLF_CRITERION = ["gini", "entropy"]
    UNDERSAMPLE_OPTIONS = ["None", "random undersample"]
    OVERSAMPLE_OPTIONS = ["None", "SMOTE", "SMOTEENN"]
    CLASS_WEIGHT_OPTIONS = ["None", "balanced", "balanced_subsample", "custom"]
    BUCKET_METHODS = [
        "fd",
        "doane",
        "auto",
        "scott",
        "stone",
        "rice",
        "sturges",
        "sqrt",
    ]
    CLF_TEST_SIZE_OPTIONS = [
        "0.1",
        "0.2",
        "0.3",
        "0.4",
        "0.5",
        "0.6",
        "0.7",
        "0.8",
        "0.9",
    ]
    PALETTE_OPTIONS = [
        "magma",
        "jet",
        "inferno",
        "plasma",
        "viridis",
        "gnuplot2",
        "RdBu",
        "winter",
    ]
    PALETTE_OPTIONS_CATEGORICAL = [
        "Pastel1",
        "Paired",
        "Accent",
        "Dark",
        "Set1",
        "Set2",
        "tab10",
    ]
    RESOLUTION_OPTIONS = [
        "320×240",
        "640×480",
        "720×480",
        "800×640",
        "960×800",
        "1120×960",
        "1280×720",
        "1980×1080",
    ]
    DPI_OPTIONS = [100, 200, 400, 800, 1600, 3200]
    SPEED_OPTIONS = [round(x, 1) for x in list(np.arange(0.1, 2.1, 0.1))]
    PERFORM_FLAGS = ["yes", True, "True"]
    RUN_OPTIONS_FLAGS = ["yes", True, "True", "False", "no", False, "true", "false"]
    SCALER_NAMES = ["MIN-MAX", "STANDARD", "QUANTILE"]
    HEATMAP_SHADING_OPTIONS = ["gouraud", "flat"]
    HEATMAP_BIN_SIZE_OPTIONS = [
        "10×10",
        "20×20",
        "40×40",
        "80×80",
        "100×100",
        "160×160",
        "320×320",
        "640×640",
        "1280×1280",
    ]
    INTERPOLATION_OPTIONS = [
        "Animal(s): Nearest",
        "Animal(s): Linear",
        "Animal(s): Quadratic",
        "Body-parts: Nearest",
        "Body-parts: Linear",
        "Body-parts: Quadratic",
    ]
    INTERPOLATION_OPTIONS_W_NONE = [
        "None",
        "Animal(s): Nearest",
        "Animal(s): Linear",
        "Animal(s): Quadratic",
        "Body-parts: Nearest",
        "Body-parts: Linear",
        "Body-parts: Quadratic",
    ]
    IMPORT_TYPE_OPTIONS = [
        "CSV (DLC/DeepPoseKit)",
        "JSON (BENTO)",
        "H5 (multi-animal DLC)",
        "SLP (SLEAP)",
        "CSV (SLEAP)",
        "H5 (SLEAP)",
        "TRK (multi-animal APT)",
        "MAT (DANNCE 3D)",
    ]
    SMOOTHING_OPTIONS = ["Gaussian", "Savitzky Golay"]
    MULTI_DLC_TYPE_IMPORT_OPTION = ["skeleton", "box", "ellipse"]
    FEATURE_SUBSET_OPTIONS = [
        "Two-point body-part distances (mm)",
        "Within-animal three-point body-part angles (degrees)",
        "Within-animal three-point convex hull perimeters (mm)",
        "Within-animal four-point convex hull perimeters (mm)",
        "Entire animal convex hull perimeters (mm)",
        "Entire animal convex hull area (mm2)",
        "Frame-by-frame body-part movements (mm)",
        "Frame-by-frame body-part distances to ROI centers (mm)",
        "Frame-by-frame body-parts inside ROIs (Boolean)",
    ]
    SMOOTHING_OPTIONS_W_NONE = ["None", "Gaussian", "Savitzky Golay"]
    VIDEO_FORMAT_OPTIONS = ["mp4", "avi"]
    ALL_VIDEO_FORMAT_OPTIONS = (".avi", ".mp4", ".mov", ".flv", ".m4v")
    ALL_IMAGE_FORMAT_OPTIONS = (".bmp", ".png", ".jpeg", ".jpg")
    ALL_IMAGE_FORMAT_STR_OPTIONS = ".bmp .png .jpeg .jpg"
    ALL_VIDEO_FORMAT_STR_OPTIONS = ".avi .mp4 .mov .flv .m4v"
    WORKFLOW_FILE_TYPE_OPTIONS = ["csv", "parquet"]
    WORKFLOW_FILE_TYPE_STR_OPTIONS = ".csv .parquet"
    TRACKING_TYPE_OPTIONS = ["Classic tracking", "Multi tracking", "3D tracking"]
    UNSUPERVISED_FEATURE_OPTIONS = [
        "INCLUDE FEATURE DATA (ORIGINAL)",
        "INCLUDE FEATURES (SCALED)",
        "EXCLUDE FEATURE DATA",
    ]
    TIMEBINS_MEASURMENT_OPTIONS = [
        "First occurrence (s)",
        "Event count",
        "Total event duration (s)",
        "Mean event duration (s)",
        "Median event duration (s)",
        "Mean event interval (s)",
        "Median event interval (s)",
    ]
    CLF_DESCRIPTIVES_OPTIONS = [
        "Bout count",
        "Total event duration (s)",
        "Mean event bout duration (s)",
        "Median event bout duration (s)",
        "First event occurrence (s)",
        "Mean event bout interval duration (s)",
        "Median event bout interval duration (s)",
    ]
    CLASSICAL_TRACKING_OPTIONS = [
        "1 animal; 4 body-parts",
        "1 animal; 7 body-parts",
        "1 animal; 8 body-parts",
        "1 animal; 9 body-parts",
        "2 animals; 8 body-parts",
        "2 animals; 14 body-parts",
        "2 animals; 16 body-parts",
        "MARS",
    ]
    MULTI_ANIMAL_TRACKING_OPTIONS = [
        "Multi-animals; 4 body-parts",
        "Multi-animals; 7 body-parts",
        "Multi-animals; 8 body-parts",
    ]
    THREE_DIM_TRACKING_OPTIONS = ["3D tracking"]
    TRAIN_TEST_SPLIT = ["FRAMES", "BOUTS"]
    BOOL_STR_OPTIONS = ["TRUE", "FALSE"]
    GANTT_VALIDATION_OPTIONS = [
        "None",
        "Gantt chart: final frame only (slightly faster)",
        "Gantt chart: video",
    ]
    THIRD_PARTY_ANNOTATION_APPS_OPTIONS = [
        "BORIS",
        "ETHOVISION",
        "OBSERVER",
        "SOLOMON",
        "DEEPETHOGRAM",
        "BENTO",
    ]
    THIRD_PARTY_ANNOTATION_ERROR_OPTIONS = [
        "INVALID annotations file data format",
        "ADDITIONAL third-party behavior detected",
        "Annotations OVERLAP conflict",
        "ZERO third-party video behavior annotations found",
        "Annotations and pose FRAME COUNT conflict",
        "Annotations EVENT COUNT conflict",
        "Annotations data file NOT FOUND",
    ]


class TextOptions(Enum):
    FIRST_LINE_SPACING = (
        2  # DISTANCE MULTIPLIER BETWEEN FIRST PRINTED ROW AND THE TOP OF THE IMAGE.
    )
    LINE_SPACING = 1  # DISTANCE MULTIPLIER BETWEEN 2nd AND THIRD, THIRD AND FOURTH ETC. PRINTED ROWS.
    BORDER_BUFFER_X = 5  # ADDITIONAL PIXEL BUFFER DISTANCE BETWEEN THE START OF THE IMAGE AND FIRST PRINTED CHARACTER ON X AXIS
    BORDER_BUFFER_Y = 10  # ADDITIONAL PIXEL BUFFER DISTANCE BETWEEN THE START OF THE IMAGE AND FIRST PRINTED CHARACTER ON Y AXIS
    FONT_SCALER = 0.8  # CONSTANT USED TO SCALE FONT ACCORDING TO RESOLUTION. INCREASING VALUE WILL RESULT IN LARGER TEXT.
    RESOLUTION_SCALER = 1500  # CONSTANT USED TO SCALE SPACINGS, FONT, AND RADIUS OF CIRCLES. LARGER NUMBER WILL RESULT IN SMALLER SPACINGS, FONT, AND RADIUS OF CIRCLES.
    RADIUS_SCALER = 10  # CONSTANT USED TO SCALE CIRCLES. INCREASING VALUE WILL RESULT IN LARGER CIRCLES.
    SPACE_SCALER = 25  # CONSTANT USED TO SCALE SPACE BETWEEN PRINTED ROWS. INCREASING VALUE WILL RESULT IN LARGER PIXEL DISTANCES BETWEEN SEQUENTIAL ROWS.
    TEXT_THICKNESS = 1  # THE THICKNESS OR "BOLDNESS" OF THE FONT IN PIXELS.
    LINE_THICKNESS = (
        2  # THICKNESS OF LINES IN CIRCLES, BOUNDING BOXES AND OTHER GEOMETRIES
    )
    COLOR = (
        147,
        20,
        255,
    )  # THE COLOR OF THE TEXT IN BGR. (147, 20, 255)  REPRESENT DEEP PINK: TYPICALLY NOT A COLOR IN ANIMAL ARENA.
    FONT = cv2.FONT_HERSHEY_SIMPLEX


class Defaults(Enum):
    MAX_TASK_PER_CHILD = 10
    LARGE_MAX_TASK_PER_CHILD = 1000
    CHUNK_SIZE = 1
    SPLASH_TIME = 2500
    try:
        WELCOME_MSG = f'Welcome fellow scientists! \n SimBA v.{pkg_resources.get_distribution("simba-uw-tf-dev").version} \n '
    except pkg_resources.DistributionNotFound:
        WELCOME_MSG = f'Welcome fellow scientists! \n SimBA v. "dev" \n '
    BROWSE_FOLDER_BTN_TEXT = "Browse Folder"
    BROWSE_FILE_BTN_TEXT = "Browse File"
    NO_FILE_SELECTED_TEXT = "No file selected"
    STR_SPLIT_DELIMITER = "\t"


class TagNames(Enum):
    GREETING = "greeting"
    COMPLETE = "complete"
    WARNING = "warning"
    ERROR = "error"
    TRASH = "trash"
    STANDARD = "standard"
    CLASS_INIT = "CLASS_INIT"


class DirNames(Enum):
    PROJECT = "project_folder"
    MODEL = "models"
    CONFIGS = "configs"
    CSV = "csv"
    FRAMES = "frames"
    LOGS = "logs"
    MEASURES = "measures"
    POSE_CONFIGS = "pose_configs"
    BP_NAMES = "bp_names"
    VIDEOS = "videos"
    FEATURES_EXTRACTED = "features_extracted"
    INPUT_CSV = "input_csv"
    MACHINE_RESULTS = "machine_results"
    OUTLIER_MOVEMENT = "outlier_corrected_movement"
    OUTLIER_MOVEMENT_LOCATION = "outlier_corrected_movement_location"
    TARGETS_INSERTED = "targets_inserted"
    INPUT = "input"
    OUTPUT = "output"


class Keys(Enum):
    ROI_RECTANGLES = "rectangles"
    ROI_CIRCLES = "circleDf"
    ROI_POLYGONS = "polygons"
    DOCUMENTATION = "documentation"
    FRAME_COUNT = "frame_count"


class Dtypes(Enum):
    NAN = "NaN"
    STR = "str"
    INT = "int"
    FLOAT = "float"
    FOLDER = "folder_path"
    NONE = "None"
    SQRT = "sqrt"
    ENTROPY = "entropy"


class Methods(Enum):
    USER_DEFINED = "user_defined"
    CLASSIC_TRACKING = "Classic tracking"
    THREE_D_TRACKING = "3D tracking"
    MULTI_TRACKING = "Multi tracking"
    CREATE_POSE_CONFIG = "Create pose config..."
    RANDOM_UNDERSAMPLE = "random undersample"
    SMOTE = "SMOTE"
    SMOTEENN = "SMOTEENN"
    GAUSSIAN = "Gaussian"
    SAVITZKY_GOLAY = "Savitzky Golay"
    SPLIT_TYPE_FRAMES = "FRAMES"
    SPLIT_TYPE_BOUTS = "BOUTS"
    BORIS = "BORIS"
    WARNING = "WARNING"
    ERROR = "ERROR"
    ANOVA = "ANOVA"
    INVALID_THIRD_PARTY_APPENDER_FILE = "INVALID annotations file data format"
    ADDITIONAL_THIRD_PARTY_CLFS = "ADDITIONAL third-party behavior detected"
    ZERO_THIRD_PARTY_VIDEO_ANNOTATIONS = "ZERO third-party video annotations found"
    THIRD_PARTY_FPS_CONFLICT = "Annotations and pose FPS conflict"
    THIRD_PARTY_EVENT_COUNT_CONFLICT = "Annotations EVENT COUNT conflict"
    THIRD_PARTY_EVENT_OVERLAP = "Annotations OVERLAP inaccuracy"
    ZERO_THIRD_PARTY_VIDEO_BEHAVIOR_ANNOTATIONS = (
        "ZERO third-party video behavior annotations found"
    )
    THIRD_PARTY_FRAME_COUNT_CONFLICT = "Annotations and pose FRAME COUNT conflict"
    THIRD_PARTY_ANNOTATION_FILE_NOT_FOUND = "Annotations data file NOT FOUND"


class OS(Enum):
    WINDOWS = "Windows"
    LINUX = "Linux"
    MAC = "Darwin"
    PYTHON_VER = str(f"{sys.version_info.major}.{sys.version_info.minor}")


class Links(Enum):
    FEATURE_SUBSETS = (
        "https://github.com/sgoldenlab/simba/blob/master/docs/feature_subsets.md"
    )
    HEATMAP_LOCATION = (
        "https://github.com/sgoldenlab/simba/blob/master/docs/ROI_tutorial.md#heatmaps"
    )
    HEATMAP_CLF = "https://github.com/sgoldenlab/simba/blob/master/docs/Scenario2.md#visualizing-classification-heatmaps"
    DATA_TABLES = "https://github.com/sgoldenlab/simba/blob/master/docs/Scenario2.md#visualizing-data-tables"
    CONCAT_VIDEOS = "https://github.com/sgoldenlab/simba/blob/master/docs/Scenario2.md#merging-concatenating-videos"
    DISTANCE_PLOTS = "https://github.com/sgoldenlab/simba/blob/master/docs/Scenario2.md#visualizing-distance-plots"
    PATH_PLOTS = "https://github.com/sgoldenlab/simba/blob/master/docs/Scenario2.md#visualizing-path-plots"
    VISUALIZE_CLF_PROBABILITIES = "https://github.com/sgoldenlab/simba/blob/master/docs/Scenario2.md#visualizing-classification-probabilities"
    GANTT_PLOTS = "https://github.com/sgoldenlab/simba/blob/master/docs/Scenario2.md#visualizing-gantt-charts"
    SKLEARN_PLOTS = "https://github.com/sgoldenlab/simba/blob/master/docs/Scenario2.md#visualizing-classifications"
    ANALYZE_ML_RESULTS = "https://github.com/sgoldenlab/simba/blob/master/docs/Scenario2.md#part-4--analyze-machine-results"
    FSTTC = "https://github.com/sgoldenlab/simba/blob/master/docs/FSTTC.md"
    KLEINBERG = (
        "https://github.com/sgoldenlab/simba/blob/master/docs/kleinberg_filter.md"
    )
    DOWNSAMPLE = "https://github.com/sgoldenlab/simba/blob/master/docs/Tutorial_tools.md#downsample-video"
    VIDEO_TOOLS = (
        "https://github.com/sgoldenlab/simba/blob/master/docs/Tutorial_tools.md"
    )
    SET_RUN_ML_PARAMETERS = "https://github.com/sgoldenlab/simba/blob/master/docs/Scenario2.md#part-3-run-the-classifier-on-new-data"
    OULIERS = (
        "https://github.com/sgoldenlab/simba/blob/master/misc/Outlier_settings.pdf"
    )
    REMOVE_CLF = "https://github.com/sgoldenlab/simba/blob/master/docs/Scenario1.md#step-2-optional-step--import-more-dlc-tracking-data-or-videos"
    ROI_FEATURES_PLOT = "https://github.com/sgoldenlab/simba/blob/master/docs/ROI_tutorial.md#part-5-visualizing-roi-features"
    ROI_DATA_PLOT = "https://github.com/sgoldenlab/simba/blob/master/docs/ROI_tutorial.md#part-4-visualizing-roi-data"
    DIRECTING_ANIMALS_PLOTS = "https://github.com/sgoldenlab/simba/blob/master/docs/directionality_between_animals.md"
    CLF_VALIDATION = (
        "https://github.com/sgoldenlab/simba/blob/master/docs/classifier_validation.md"
    )
    BATCH_PREPROCESS = "https://github.com/sgoldenlab/simba/blob/master/docs/tutorial_process_videos.md"
    THIRD_PARTY_ANNOTATION_NEW = (
        "https://github.com/sgoldenlab/simba/blob/master/docs/third_party_annot_new.md"
    )
    OUT_OF_SAMPLE_VALIDATION = "https://github.com/sgoldenlab/simba/blob/master/docs/Scenario1.md#step-8-evaluating-the-model-on-new-out-of-sample-data"
    ROI = "https://github.com/sgoldenlab/simba/blob/master/docs/ROI_tutorial_new.md"
    ROI_FEATURES = "https://github.com/sgoldenlab/simba/blob/master/docs/ROI_tutorial.md#part-3-generating-features-from-roi-data"
    ROI_DATA_ANALYSIS = "https://github.com/sgoldenlab/simba/blob/master/docs/ROI_tutorial.md#part-2-analyzing-roi-data"
    DATA_ANALYSIS = "https://github.com/sgoldenlab/simba/blob/master/docs/Scenario2.md#part-4--analyze-machine-results"
    ANALYZE_ROI = "https://github.com/sgoldenlab/simba/blob/master/docs/ROI_tutorial.md#part-2-analyzing-roi-data"
    EXTRACT_FEATURES = "https://github.com/sgoldenlab/simba/blob/master/docs/Scenario1.md#step-5-extract-features"
    USER_DEFINED_FEATURE_EXTRACTION = (
        "https://github.com/sgoldenlab/simba/blob/master/docs/extractFeatures.md"
    )
    APPEND_ROI_FEATURES = "https://github.com/sgoldenlab/simba/blob/master/docs/ROI_tutorial.md#part-3-generating-features-from-roi-data"
    LABEL_BEHAVIOR = (
        "https://github.com/sgoldenlab/simba/blob/master/docs/label_behavior.md"
    )
    THIRD_PARTY_ANNOTATION = (
        "https://github.com/sgoldenlab/simba/blob/master/docs/third_party_annot.md"
    )
    PSEUDO_LBL = "https://github.com/sgoldenlab/simba/blob/master/docs/pseudoLabel.md"
    ADVANCED_LBL = (
        "https://github.com/sgoldenlab/simba/blob/master/docs/advanced_labelling.md"
    )
    TRAIN_ML_MODEL = "https://github.com/sgoldenlab/simba/blob/master/docs/Scenario1.md#step-7-train-machine-model"
    VISUALIZATION = "https://github.com/sgoldenlab/simba/blob/master/docs/Scenario2.md#part-5--visualizing-results"
    PLOTLY = "https://github.com/sgoldenlab/simba/blob/master/docs/plotly_dash.md"
    VIDEO_PARAMETERS = "https://github.com/sgoldenlab/simba/blob/master/docs/Scenario1.md#step-3-set-video-parameters"
    OUTLIERS_DOC = "https://github.com/sgoldenlab/simba/blob/master/docs/Scenario1.md#step-4-outlier-correction"
    CREATE_PROJECT = "https://github.com/sgoldenlab/simba/blob/master/docs/Scenario1.md#step-1-generate-project-config"
    LOAD_PROJECT = "https://github.com/sgoldenlab/simba/blob/master/docs/Scenario1.md#part-2-load-project-1"
    SCENARIO_2 = "https://github.com/sgoldenlab/simba/blob/master/docs/Scenario2.md"
    BBOXES = "https://github.com/sgoldenlab/simba/blob/master/docs/anchored_rois.md"
    CUE_LIGHTS = (
        "https://github.com/sgoldenlab/simba/blob/master/docs/cue_light_tutorial.md"
    )
    ADDITIONAL_IMPORTS = "https://github.com/sgoldenlab/simba/blob/master/docs/Scenario1.md#step-2-optional-step--import-more-dlc-tracking-data-or-videos"
    AGGREGATE_BOOL_STATS = "https://github.com/sgoldenlab/simba/blob/master/docs/ROI_tutorial.md#compute-aggregate-conditional-statistics-from-boolean-fields"
    GITHUB_REPO = "https://github.com/sgoldenlab/simba"
    OSF_REPO = "https://osf.io/tmu6y/"
    GITTER = "https://gitter.im/SimBA-Resource/community"
    COUNT_ANNOTATIONS_IN_PROJECT = "https://github.com/sgoldenlab/simba/blob/master/docs/label_behavior.md#count-annotations-in-simba-project"
    COUNT_ANNOTATIONS_OUTSIDE_PROJECT = "https://github.com/sgoldenlab/simba/blob/master/docs/Tutorial_tools.md#extract-project-annotation-counts"


class Labelling(Enum):
    PADDING = 5
    MAX_FRM_SIZE = (1280, 650)
    VIDEO_FRAME_SIZE = (700, 500)
    PLAY_VIDEO_SCRIPT_PATH = os.path.join(
        os.path.dirname(simba.__file__), "labelling", "play_annotation_video.py"
    )
    VALID_ANNOTATIONS_ADVANCED = [0, 1, 2]


class GeometryEnum(Enum):
    CAP_STYLE_MAP = {"round": 1, "square": 2, "flat": 3}
    HISTOGRAM_COMPARISON_MAP = {
        "correlation": 0,
        "chi_square": 1,
        "intersection": 2,
        "bhattacharyya": 3,
        "hellinger": 4,
        "chi_square_alternative": 5,
    }
    CONTOURS_MODE_MAP = {"exterior": 0, "all": 1, "interior": 3}
    CONTOURS_RETRIEVAL_MAP = {"simple": 2, "none": 0, "l1": 3, "kcos": 4}
    RANKING_METHODS = [
        "area",
        "min_distance",
        "max_distance",
        "mean_distance",
        "left_to_right",
        "top_to_bottom",
    ]


class MLParamKeys(Enum):
    CLASSIFIER = "classifier"
    RF_ESTIMATORS = "rf_n_estimators"
    RF_CRITERION = "rf_criterion"
    TT_SIZE = "train_test_size"
    MIN_LEAF = "rf_min_sample_leaf"
    RF_METADATA = "generate_rf_model_meta_data_file"
    EX_DECISION_TREE = "generate_example_decision_tree"
    CLF_REPORT = "generate_classification_report"
    IMPORTANCE_LOG = "generate_features_importance_log"
    IMPORTANCE_BAR_CHART = "generate_features_importance_bar_graph"
    PERMUTATION_IMPORTANCE = "compute_feature_permutation_importance"
    LEARNING_CURVE = "generate_sklearn_learning_curves"
    PRECISION_RECALL = "generate_precision_recall_curves"
    RF_MAX_FEATURES = "rf_max_features"
    RF_MAX_DEPTH = "rf_max_depth"
    LEARNING_CURVE_K_SPLITS = "learning_curve_k_splits"
    LEARNING_CURVE_DATA_SPLITS = "learning_curve_data_splits"
    N_FEATURE_IMPORTANCE_BARS = "n_feature_importance_bars"
    SHAP_SCORES = "generate_shap_scores"
    SHAP_PRESENT = "shap_target_present_no"
    SHAP_ABSENT = "shap_target_absent_no"
    SHAP_SAVE_ITERATION = "shap_save_iteration"
    PARTIAL_DEPENDENCY = "partial_dependency"
    TRAIN_TEST_SPLIT_TYPE = "train_test_split_type"
    UNDERSAMPLE_SETTING = "under_sample_setting"
    UNDERSAMPLE_RATIO = "under_sample_ratio"
    OVERSAMPLE_SETTING = "over_sample_setting"
    OVERSAMPLE_RATIO = "over_sample_ratio"
    CLASS_WEIGHTS = "class_weights"
    CLASS_CUSTOM_WEIGHTS = "class_custom_weights"
    EX_DECISION_TREE_FANCY = "generate_example_decision_tree_fancy"
    IMPORTANCE_BARS_N = "N_feature_importance_bars"
    LEARNING_DATA_SPLITS = "LearningCurve_shuffle_data_splits"
    MODEL_TO_RUN = "model_to_run"
    SAVE_TRAIN_TEST_FRM_IDX = "save_train_test_frm_idx"
    SHAP_MULTIPROCESS = "shap_multiprocess"
    CLASSIFIER_MAP = "classifier_map"


class TestPaths(Enum):
    CRITICAL_VALUES = "../simba/assets/lookups/critical_values_05.pickle"



__author__ = "Simon Nilsson"

from tkinter import messagebox as mb

from simba.utils.enums import Defaults, TagNames
from simba.utils.printing import log_event

WINDOW_TITLE = "SIMBA ERROR"


class SimbaError(Exception):
    def __init__(self, msg: str, source: str = " ", show_window: bool = False):
        self.msg, self.source, self.show_window = msg, source, show_window
        self.print_and_log_error()

    def __str__(self):
        return self.msg

    def print_and_log_error(self):
        log_event(
            logger_name=f"{self.source}.{self.__class__.__name__}",
            log_type=TagNames.ERROR.value,
            msg=self.msg,
        )
        print(f"{self.msg}{Defaults.STR_SPLIT_DELIMITER.value}{TagNames.ERROR.value}")
        if self.show_window:
            mb.showerror(title=WINDOW_TITLE, message=self.msg)


class NoSpecifiedOutputError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = True):
        super().__init__(msg=msg, source=source, show_window=show_window)


class ROICoordinatesNotFoundError(SimbaError):
    def __init__(
        self, expected_file_path: str, source: str = "", show_window: bool = False
    ):
        msg = f"SIMBA ROI COORDINATES ERROR: No ROI coordinates found. Please use the [ROI] tab to define ROIs. Expected at location {expected_file_path}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class NoChoosenClassifierError(SimbaError):
    def __init__(self, source: str = "", show_window: bool = False):
        msg = f"Select at least one classifier"
        super().__init__(msg=msg, source=source, show_window=show_window)


class NoChoosenROIError(SimbaError):
    def __init__(self, source: str = "", show_window: bool = False):
        msg = f"Please select at least one ROI."
        super().__init__(msg=msg, source=source, show_window=show_window)


class NoChoosenMeasurementError(SimbaError):
    def __init__(self, source: str = "", show_window: bool = False):
        msg = "SIMBA NoChoosenMeasurementError ERROR: Please select at least one measurement to calculate descriptive statistics for."
        super().__init__(msg=msg, source=source, show_window=show_window)


class NoDataError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA NO DATA ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class SamplingError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA SAMPLING ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class PermissionError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA PERMISSION ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class NoROIDataError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA NO ROI DATA ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class MixedMosaicError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA MixedMosaicError ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class ClassifierInferenceError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA CLASSIFIER INFERENCE ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class AnimalNumberError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA ANIMAL NUMBER ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class InvalidFilepathError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA INVALID FILE PATH ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class NoFilesFoundError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA NO FILES FOUND ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class DataHeaderError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA DATA HEADER ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class NotDirectoryError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA NOT A DIRECTORY ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class DirectoryExistError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA DIRECTORY ALREADY EXIST ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class FileExistError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA FILE EXIST ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class FrameRangeError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA FRAME RANGE ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class AdvancedLabellingError(SimbaError):
    def __init__(
        self,
        frame: str,
        lbl_lst: list,
        unlabel_lst: list,
        source: str = "",
        show_window: bool = False,
    ):
        msg = (
            "SIMBA ADVANCED LABELLING ERROR: In advanced labelling of multiple behaviors, any annotated frame cannot have some "
            "behaviors annotated as present/absent, while other behaviors are un-labelled. All behaviors need "
            "labels for a frame with one or more labels. In frame {}, behaviors {} are labelled, while behaviors "
            "{} are un-labelled.".format(str(frame), lbl_lst, unlabel_lst)
        )
        super().__init__(msg=msg, source=source, show_window=show_window)


class InvalidHyperparametersFileError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA HYPERPARAMETER FILE ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class InvalidVideoFileError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA VIDEO FILE ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class InvalidFileTypeError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA INVALID FILE TYPE ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class FaultyTrainingSetError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA INVALID ML TRAINING SET ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class CountError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA COUNT ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class FeatureNumberMismatchError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA FEATURE NUMBER MISMATCH ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class DuplicationError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA DUPLICATION ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class InvalidInputError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA VALUE ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class IntegerError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA INTEGER ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class StringError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA STRING ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class FloatError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA FLOAT ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class MissingProjectConfigEntryError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA MISSING PROJECT CONFIG ENTRY ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class MissingColumnsError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA MISSING COLUMN ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class CorruptedFileError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA READ FILE ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class ParametersFileError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA VIDEO PARAMETERS FILE ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class ColumnNotFoundError(SimbaError):
    def __init__(
        self,
        column_name: str,
        file_name: str,
        source: str = "",
        show_window: bool = False,
    ):
        msg = f"SIMBA ERROR: Field name {column_name} could not be found in file {file_name}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class BodypartColumnNotFoundError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA BODY_PART COLUMN NOT FOUND ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class AnnotationFileNotFoundError(SimbaError):
    def __init__(self, video_name: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA THIRD-PARTY ANNOTATION ERROR: NO ANNOTATION DATA FOR VIDEO {video_name} FOUND"
        super().__init__(msg=msg, source=source, show_window=show_window)


class DirectoryNotEmptyError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA DIRECTORY NOT EMPTY ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


#####


class ThirdPartyAnnotationFileNotFoundError(SimbaError):
    def __init__(self, video_name: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA ERROR: Could not find file in project_folder/csv/features_extracted directory representing annotations for video {video_name}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class ThirdPartyAnnotationsFpsConflictError(SimbaError):
    def __init__(
        self,
        video_name: str,
        annotation_fps: int,
        video_fps: int,
        source: str = "",
        show_window: bool = False,
    ):
        msg = f"SIMBA THIRD-PARTY ANNOTATION ERROR: The FPS for video {video_name} is set to {str(video_fps)} in SimBA and {str(annotation_fps)} in the annotation file"
        super().__init__(msg=msg, source=source, show_window=show_window)


class ThirdPartyAnnotationsMissingAnnotationsError(SimbaError):
    def __init__(
        self,
        video_name: str,
        clf_names: list,
        source: str = "",
        show_window: bool = False,
    ):
        msg = f"SIMBA THIRD-PARTY ANNOTATION ERROR: No annotations detected for SimBA classifier(s) named {clf_names} for video {video_name}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class ThirdPartyAnnotationOverlapError(SimbaError):
    def __init__(
        self,
        video_name: str,
        clf_name: str,
        source: str = "",
        show_window: bool = False,
    ):
        msg = f"SIMBA THIRD-PARTY ANNOTATION ERROR: The annotations for behavior {clf_name} in video {video_name} contains behavior start events that are initiated PRIOR to the PRECEDING behavior event ending. SimBA requires a specific behavior event to end before another behavior event can start."
        super().__init__(msg=msg, source=source, show_window=show_window)


class ThirdPartyAnnotationsAdditionalClfError(SimbaError):
    def __init__(
        self,
        video_name: str,
        clf_names: list,
        source: str = "",
        show_window: bool = False,
    ):
        msg = f"SIMBA THIRD-PARTY ANNOTATION ERROR: Annotations file for video {video_name} has annotations for the following behaviors {clf_names} that are NOT classifiers named in the SimBA project."
        super().__init__(msg=msg, source=source, show_window=show_window)


class ThirdPartyAnnotationEventCountError(SimbaError):
    def __init__(
        self,
        video_name: str,
        clf_name: str,
        start_event_cnt: int,
        stop_event_cnt: int,
        source: str = "",
        show_window: bool = False,
    ):
        msg = f"SIMBA THIRD-PARTY ANNOTATION ERROR: The annotations for behavior {clf_name} in video {video_name} contains {str(start_event_cnt)} start events and {str(stop_event_cnt)} stop events. SimBA requires the number of stop and start event counts to be equal."
        super().__init__(msg=msg, source=source, show_window=show_window)


class ThirdPartyAnnotationsClfMissingError(SimbaError):
    def __init__(
        self,
        video_name: str,
        clf_name: str,
        source: str = "",
        show_window: bool = False,
    ):
        msg = f"SIMBA THIRD-PARTY ANNOTATION WARNING: No annotations detected for video {video_name} and behavior {clf_name}."
        super().__init__(msg=msg, source=source, show_window=show_window)


class ThirdPartyAnnotationsOutsidePoseEstimationDataError(SimbaError):
    def __init__(
        self,
        video_name: str,
        frm_cnt: int,
        clf_name: str or None = None,
        annotation_frms: int or None = None,
        first_error_frm: int or None = None,
        ambiguous_cnt: int or None = None,
        source: str = "",
        show_window: bool = False,
    ):

        if clf_name:
            msg = (
                f"SIMBA THIRD-PARTY ANNOTATION WARNING: SimBA found THIRD-PARTY annotations for behavior {clf_name} in video "
                f"{video_name} that are annotated to occur at times which is not present in the "
                f"video data you imported into SIMBA. The video you imported to SimBA has {str(frm_cnt)} frames. "
                f"However, in BORIS, you have annotated {clf_name} to happen at frame number {str(first_error_frm)}. "
                f"These ambiguous annotations occur in {str(ambiguous_cnt)} different frames for video {video_name} that SimBA will **remove** by default. "
                f"Please make sure you imported the same video as you annotated in BORIS into SimBA and the video is registered with the correct frame rate."
            )
        else:
            msg = f"SIMBA THIRD-PARTY ANNOTATION WARNING: The annotations for video {video_name} contain data for {str(annotation_frms)} frames. The pose-estimation features for the same video contain data for {str(frm_cnt)} frames."

        super().__init__(msg=msg, source=source, show_window=show_window)


class FFMPEGCodecGPUError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA FFMPEG CODEC ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class FFMPEGNotFoundError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA FFMPEG NOT FOUND ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


class ArrayError(SimbaError):
    def __init__(self, msg: str, source: str = "", show_window: bool = False):
        msg = f"SIMBA ARRAY SIZE ERROR: {msg}"
        super().__init__(msg=msg, source=source, show_window=show_window)


# test = NoSpecifiedOutputError(msg='test', source='test.method')
# test = FFMPEGNotFoundError(msg='323')

# class NoSpecifiedOutputError(SimbaError):
#     @log_error_decorator()
#     def __init__(self, msg: str, source: str = '', show_window: bool = True):



__author__ = "Simon Nilsson"

import itertools
import os
from typing import List, Optional, Union

import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import check_if_filepath_list_is_empty
from simba.utils.data import detect_bouts
from simba.utils.enums import TagNames
from simba.utils.errors import CountError
from simba.utils.printing import log_event, stdout_success
from simba.utils.read_write import get_fn_ext, read_df


class FSTTCCalculator(ConfigReader, PlottingMixin):
    """
    Compute forward spike-time tiling coefficients between pairs of
    classified behaviors.

    :parameter str config_path: path to SimBA project config file in Configparser format.
    :parameter: Optional[bool] join_bouts_within_delta: If several bouts onsets (of the same classifier) occurs within a single time-delta, then join the bouts into a single bout.
    :parameter: Optional[bool] time_delta_at_onset: If True, time delta is initatiated at bout onset. If False, then initated at bout offset and includes bout duration. Default: False.
    :parameter int time_window: FSTTC hyperparameter; Integer representing the time window in seconds.
    :parameter List[str] behavior_lst: Behaviors to calculate FSTTC between. FSTTC will be computed for all combinations of behaviors.
    :parameter bool create_graphs: If True, created violin plots (as below) representing each FSTTC. Default: False.

    .. image:: _static/img/fsttc_violin.png
       :width: 500
       :align: center

    .. note::
       `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/FSTTC.md>`__.

    Examples
    -----
    >>> fsttc_calculator = FSTTCCalculator(config_path='MyConfigPath', time_window=2, behavior_lst=['Attack', 'Sniffing'], create_graphs=True)
    >>> fsttc_calculator.run()

    References
    ----------
    .. [1] Lee et al., Temporal microstructure of dyadic social behavior during relationship formation in mice, `PLOS One`,
           2019.
    .. [2] Cutts et al., Detecting Pairwise Correlations in Spike Trains: An Objective Comparison of Methods and
           Application to the Study of Retinal Waves, `J Neurosci`, 2014.
    """

    def __init__(
        self,
        config_path: Union[str, os.PathLike],
        time_window: int,
        behavior_lst: List[str],
        time_delta_at_onset: Optional[bool] = False,
        join_bouts_within_delta: Optional[bool] = False,
        create_graphs: Optional[bool] = False,
    ):
        ConfigReader.__init__(self, config_path=config_path)
        PlottingMixin.__init__(self)
        log_event(
            logger_name=str(self.__class__.__name__),
            log_type=TagNames.CLASS_INIT.value,
            msg=self.create_log_msg_from_init_args(locals=locals()),
        )
        self.time_delta = int(time_window)
        self.behavior_lst = behavior_lst
        if len(self.behavior_lst) < 2:
            raise CountError(
                msg="FSTCC requires at least two behaviors",
                source=self.__class__.__name__,
            )
        self.graph_status, self.join_bouts_within_delta, self.time_delta_at_onset = (
            create_graphs,
            join_bouts_within_delta,
            time_delta_at_onset,
        )
        check_if_filepath_list_is_empty(
            filepaths=self.machine_results_paths,
            error_msg=f"Cannot calculate FSTTC, no data found in {self.machine_results_paths} directory",
        )
        self.clf_permutations = list(itertools.permutations(self.behavior_lst, 2))
        print(f"Processing FSTTC for {str(len(self.machine_results_paths))} file(s)...")

    def __join_bouts(self):
        results = []
        if self.time_delta_at_onset:
            self.bouts_df["DELTA_END"] = pd.to_datetime(
                self.bouts_df["Start_frame"] + self.frames_in_window
            )
        else:
            self.bouts_df["DELTA_END"] = pd.to_datetime(
                self.bouts_df["End_frame"] + self.frames_in_window
            )
        self.bouts_df["Start_frame"] = pd.to_datetime(self.bouts_df["Start_frame"])

        for clf in self.bouts_df["Event"].unique():
            clf_df = self.bouts_df[self.bouts_df["Event"] == clf].reset_index(drop=True)
            grouped_clf = (
                (
                    clf_df.sort_values(by=["Start_frame", "End_frame", "DELTA_END"])
                    .assign(
                        max_End=lambda d: d["DELTA_END"].cummax(),
                        group=lambda d: d["Start_frame"]
                        .ge(d["max_End"].shift())
                        .cumsum(),
                    )
                    .groupby("group")
                    .agg({"Start_frame": "min", "DELTA_END": "max"})
                    .assign(Duration=lambda g: g["DELTA_END"] - g["Start_frame"])
                )
                .drop("Duration", axis=1)
                .reset_index(drop=True)
            )
            grouped_clf["Start_frame"] = grouped_clf["Start_frame"].astype(np.int64)
            grouped_clf["End_frame"] = grouped_clf["DELTA_END"].astype(np.int64)
            grouped_clf["Event"] = clf
            results.append(grouped_clf)
        return (
            pd.concat(results, axis=0)
            .sort_values(by=["Event", "Start_frame"])
            .drop("DELTA_END", axis=1)
        )

    def find_sequences(self):
        """
        Method to create list of dataframes holding information on the sequences of behaviors including
        inter-temporal distances.

        Returns
        -------
        Attribute: list
            vide_df_sequence_lst
        """

        self.video_sequences = {}
        out_columns = [
            "Video",
            "First behaviour",
            "First behaviour start frame",
            "First behavior end frame",
            "Second behaviour",
            "Second behaviour start frame",
            "Difference: first behavior start to second behavior start",
            "Time 2nd behaviour start to time window end",
        ]

        for file_cnt, file_path in enumerate(self.machine_results_paths):
            _, self.video_name, _ = get_fn_ext(file_path)
            self.video_sequences[self.video_name] = {}
            print(
                f"Analyzing behavioral sequences: {self.video_name} (Video {file_cnt + 1}/{len(self.machine_results_paths)})..."
            )
            _, _, self.fps = self.read_video_info(video_name=self.video_name)
            self.video_sequences[self.video_name]["fps"] = self.fps
            self.frames_in_window = int((self.fps / 1000) * self.time_delta)
            self.data_df = read_df(file_path, self.file_type)  # [self.behavior_lst]
            self.video_sequences[self.video_name]["session_length_frames"] = len(
                self.data_df
            )
            self.bouts_df = detect_bouts(
                data_df=self.data_df, target_lst=self.behavior_lst, fps=self.fps
            )
            self.bouts_df["Start_frame"] = (
                self.bouts_df["Start_time"] * self.fps
            ).astype(int) - 1
            self.bouts_df = self.bouts_df[["Event", "Start_frame", "End_frame"]]
            if self.join_bouts_within_delta:
                self.bouts_df = self.__join_bouts()
            for first_clf, second_clf in self.clf_permutations:
                self.vide_df_sequence_lst = []
                sequence_name = "FSTTC {} {}".format(first_clf, second_clf)
                first_clf_df = (
                    self.bouts_df[self.bouts_df["Event"] == first_clf]
                    .sort_values(by=["Start_frame"])
                    .reset_index(drop=True)
                )
                second_clf_df = (
                    self.bouts_df[self.bouts_df["Event"] == second_clf]
                    .sort_values(by=["Start_frame"])
                    .reset_index(drop=True)
                )
                for index, row in first_clf_df.iterrows():
                    if self.time_delta_at_onset:
                        frame_crtrn_min, frame_crtrn_max = (
                            row["Start_frame"] + 1,
                            row["Start_frame"] + self.frames_in_window,
                        )
                    else:
                        if not self.join_bouts_within_delta:
                            frame_crtrn_min, frame_crtrn_max = (
                                row["Start_frame"] + 1,
                                row["End_frame"] + self.frames_in_window,
                            )
                        else:
                            frame_crtrn_min, frame_crtrn_max = (
                                row["Start_frame"] + 1,
                                row["End_frame"],
                            )

                    second_clf_df_crtrn = second_clf_df.loc[
                        (second_clf_df["Start_frame"] >= frame_crtrn_min)
                        & (second_clf_df["Start_frame"] <= frame_crtrn_max)
                    ]
                    if len(second_clf_df_crtrn) > 0:
                        second_clf_df_crtrn = second_clf_df_crtrn.head(1)
                        frames_between_behaviors = (
                            second_clf_df_crtrn["Start_frame"] - row["Start_frame"]
                        ).values
                        second_clf_df_crtrn["Frames_between_behaviors"] = (
                            second_clf_df_crtrn["Start_frame"] - row["Start_frame"]
                        )
                        if frames_between_behaviors <= 0:
                            frames_between_behaviors = 1
                        second_clf_df_crtrn["Frames_between_behaviors"] = (
                            frames_between_behaviors
                        )
                        if self.time_delta_at_onset:
                            second_clf_df_crtrn[
                                "Frames_between_second_behavior_start_to_time_window_end"
                            ] = (
                                row["Start_frame"] + self.frames_in_window
                            ) - second_clf_df_crtrn[
                                "Start_frame"
                            ].values[
                                0
                            ]
                        else:
                            second_clf_df_crtrn[
                                "Frames_between_second_behavior_start_to_time_window_end"
                            ] = (
                                row["End_frame"] + self.frames_in_window
                            ) - second_clf_df_crtrn[
                                "Start_frame"
                            ].values[
                                0
                            ]
                        self.vide_df_sequence_lst.append(
                            pd.DataFrame(
                                [
                                    [
                                        self.video_name,
                                        first_clf,
                                        row["Start_frame"],
                                        row["End_frame"],
                                        second_clf,
                                        second_clf_df_crtrn["Start_frame"].values[0],
                                        second_clf_df_crtrn[
                                            "Frames_between_behaviors"
                                        ].values[0],
                                        second_clf_df_crtrn[
                                            "Frames_between_second_behavior_start_to_time_window_end"
                                        ].values[0],
                                    ]
                                ],
                                columns=out_columns,
                            )
                        )
                    else:
                        self.vide_df_sequence_lst.append(
                            pd.DataFrame(
                                [
                                    [
                                        self.video_name,
                                        first_clf,
                                        int(row["Start_frame"]),
                                        int(row["End_frame"]),
                                        "None",
                                        "None",
                                        "None",
                                        "None",
                                    ]
                                ],
                                columns=out_columns,
                            )
                        )

                if len(self.vide_df_sequence_lst) > 0:
                    video_sequences = (
                        pd.concat(self.vide_df_sequence_lst, axis=0)
                        .drop_duplicates(
                            subset=[
                                "Video",
                                "First behaviour",
                                "First behaviour start frame",
                                "First behavior end frame",
                                "Second behaviour",
                            ],
                            keep="first",
                        )
                        .reset_index(drop=True)
                    )
                    if self.time_delta_at_onset:
                        video_sequences["Total_window_frames"] = self.frames_in_window
                    else:
                        video_sequences["Total_window_frames"] = (
                            video_sequences["First behavior end frame"]
                            - video_sequences["First behaviour start frame"]
                        ) + self.frames_in_window
                    self.video_sequences[self.video_name][
                        sequence_name
                    ] = video_sequences
                else:
                    self.video_sequences[self.video_name][sequence_name] = None

    def run(self):
        """
        Method to calculate forward spike-time tiling coefficients (FSTTC) using the data computed in :meth:
        :meth:`~simba.FSTTCPerformer.find_sequences`.

        Returns
        -------
        Attribute: dict
            results_dict
        """

        self.find_sequences()
        self.results_dict = {}
        for video_name, video_data in self.video_sequences.items():
            self.results_dict[video_name] = {}
            fps, session_frames = video_data["fps"], video_data["session_length_frames"]
            for first_clf, second_clf in self.clf_permutations:
                if first_clf not in self.results_dict[video_name].keys():
                    self.results_dict[video_name][first_clf] = {}
                self.results_dict[video_name][first_clf][second_clf] = {}
                sequence_data = video_data[f"FSTTC {first_clf} {second_clf}"]
                if sequence_data is None:
                    self.results_dict[video_name][first_clf][second_clf] = "No events"
                else:
                    len_clf_1 = len(
                        sequence_data[sequence_data["First behaviour"] == first_clf]
                    )
                    len_clf_1_2 = len(
                        sequence_data[
                            (sequence_data["First behaviour"] == first_clf)
                            & (sequence_data["Second behaviour"] == second_clf)
                        ]
                    )
                    if (len_clf_1 > 0) & (len_clf_1_2 == 0):
                        self.results_dict[video_name][first_clf][second_clf] = 0.0
                    else:
                        clf_1_2_df = sequence_data[
                            (sequence_data["First behaviour"] == first_clf)
                            & (sequence_data["Second behaviour"] == second_clf)
                        ]
                        P = len_clf_1_2 / len_clf_1
                        Ta = sum(clf_1_2_df["Total_window_frames"]) / session_frames
                        Tb = (
                            sum(
                                clf_1_2_df[
                                    "Time 2nd behaviour start to time window end"
                                ]
                            )
                            / session_frames
                        )
                        self.results_dict[video_name][first_clf][second_clf] = 0.5 * (
                            (P - Tb) / (1 - (P * Tb)) + ((P - Ta) / (1 - (P * Ta)))
                        )
                        # print(self.results_dict[video_name][first_clf][second_clf])
        self.save()
        if self.graph_status:
            self.__plot_FSTTC()

    def __plot_FSTTC(self) -> None:
        """
        Private method to visualize forward spike-time tiling coefficients (FSTTC) as png violin plots. Results are stored on
        disk within the `project_folder/logs` directory.
        """

        self.out_df["BEHAVIOR COMBINATION"] = self.out_df["FIRST BEHAVIOR"].str.cat(
            self.out_df["SECOND BEHAVIOR"], sep="-"
        )
        data_df = self.out_df[self.out_df["FSTTC"] != "No events"].reset_index(
            drop=True
        )
        data_df["FSTTC"] = pd.to_numeric(data_df["FSTTC"])
        self.violin_plot(
            data=data_df,
            x="BEHAVIOR COMBINATION",
            y="FSTTC",
            save_path=os.path.join(self.logs_path, f"FSTTC_{self.datetime}.png"),
        )

    def save(self):
        """
        Method to save forward spike-time tiling coefficients (FSTTC) to disk within the `project_folder/logs` directory.

        Returns
        -------
        None
        """
        self.out_df = pd.DataFrame(
            columns=["VIDEO", "FIRST BEHAVIOR", "SECOND BEHAVIOR", "FSTTC"]
        )
        for video_name, video_data in self.results_dict.items():
            for first_behavior, first_behavior_data in video_data.items():
                for second_behavior, fsttc in first_behavior_data.items():
                    self.out_df.loc[len(self.out_df)] = [
                        video_name,
                        first_behavior,
                        second_behavior,
                        fsttc,
                    ]
        self.file_save_path = os.path.join(
            self.logs_path, "FSTTC_{}.csv".format(str(self.datetime))
        )
        self.out_df.to_csv(self.file_save_path)
        self.timer.stop_timer()
        stdout_success(
            msg=f"FSTTC data saved at {self.file_save_path}",
            elapsed_time=self.timer.elapsed_time_str,
            source=self.__class__.__name__,
        )


# test = FSTTCCalculator(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                        time_window=60000,
#                        join_bouts_within_delta=True,
#                        time_delta_at_onset=True,
#                        behavior_lst=['licking_grooming', 'active_nursing', 'nest_attendance'], #'passive_nursing', 'nest_attendance'
#                        create_graphs=True)
# test.run()


# test = FSTTCCalculator(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                      time_window=10000,
#                      behavior_lst=['Attack', 'Sniffing'],
#                     create_graphs=False)
# test.run()


#
# test = FSTTCCalculator(config_path='/Users/simon/Desktop/envs/troubleshooting/naresh/project_folder/project_config.ini',
#                      time_window=2000,
#                      behavior_lst=['Erratic Turning', 'Bottom', 'Normal Swimming', 'Freezing', 'Wall Bumping'],
#                     create_graphs=True)
# test.run()



__author__ = "Simon Nilsson"

import os.path
from typing import Dict, List

import cv2

from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import check_that_column_exist
from simba.utils.enums import Dtypes
from simba.utils.errors import FrameRangeError
from simba.utils.printing import stdout_success
from simba.utils.read_write import (find_video_of_file, get_fn_ext,
                                    get_video_meta_data, read_df)


class AnnotationFrameExtractor(ConfigReader):
    """
    Extracts all human annotated frames where behavior is annotated as present into .pngs within a SimBA project.

    :param list clfs: Names of classifiers to extract behavior-present images from.
    :param dict settings: User-defined settings. E.g., how much to downsample the png images.
    :param str config_path: path to SimBA configparser.ConfigParser project_config.in

    .. note::
        Use settings={'downsample': 2} to downsample all .pngs to 0.5 original size for disk space savings.

    :example:
    >>> extractor = AnnotationFrameExtractor(config_path='project_folder/project_config.ini', clfs=['Sniffing', 'Attack'], settings={'downsample': 2})
    >>> extractor.run()
    """

    def __init__(self, clfs: List[str], settings: Dict[str, int], config_path: str):
        self.clfs = clfs
        self.settings = settings
        super().__init__(config_path=config_path)

    def run(self):
        if not os.path.exists(self.annotated_frm_dir):
            os.makedirs(self.annotated_frm_dir)
        for file_path in self.target_file_paths:
            _, video_name, _ = get_fn_ext(filepath=file_path)
            video_path = find_video_of_file(
                video_dir=self.video_dir, filename=video_name
            )
            df = read_df(file_path=file_path, file_type=self.file_type)
            for clf in self.clfs:
                check_that_column_exist(df=df, column_name=clf, file_name=file_path)
            cap = cv2.VideoCapture(video_path)
            video_meta_data = get_video_meta_data(video_path=video_path)
            for clf in self.clfs:
                save_dir = os.path.join(self.annotated_frm_dir, video_name, clf)
                if not os.path.exists(save_dir):
                    os.makedirs(save_dir)
                annot_idx = list(df.index[df[clf] == 1])
                for frm_cnt, frm in enumerate(annot_idx):
                    cap.set(1, frm)
                    ret, img = cap.read()
                    if not ret:
                        raise FrameRangeError(
                            msg=f'Frame {str(frm+1)} is annotated as {clf} present. But frame {str(frm+1)} does not exist in video file {video_path}. The video file contains {video_meta_data["frame_count"]} frames.'
                        )
                    if self.settings["downsample"] != Dtypes.NONE.value:
                        img = cv2.resize(
                            img,
                            (
                                int(img.shape[1] / self.settings["downsample"]),
                                int((img.shape[0] / self.settings["downsample"])),
                            ),
                            cv2.INTER_NEAREST,
                        )
                    cv2.imwrite(os.path.join(save_dir, f"{str(frm)}.png"), img)
                    print(
                        f"Saved {clf} annotated img ({str(frm_cnt)}/{str(len(annot_idx))}), Video: {video_name}"
                    )
        self.timer.stop_timer()
        stdout_success(
            msg=f"Annotated frames saved in {self.annotated_frm_dir} directory",
            elapsed_time=self.timer.elapsed_time_str,
        )


# test = AnnotationFrameExtractor(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                 clfs=['Sniffing', 'Attack'],
#                                 settings={'downsample': 2})
# test.run()



__author__ = "Simon Nilsson"

import functools
import itertools
import multiprocessing
import os
import platform

import cv2

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.roi_tools.ROI_feature_analyzer import ROIFeatureCreator
from simba.utils.checks import check_file_exist_and_readable
from simba.utils.enums import Formats, Paths
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (concatenate_videos_in_folder, get_fn_ext,
                                    get_video_meta_data, read_df,
                                    read_video_info_csv, remove_a_folder)


class ROIfeatureVisualizerMultiprocess(ConfigReader, PlottingMixin):
    """
    Visualize features that depend on the relationships between the location of the animals and user-defined
    ROIs. E.g., distances to centroids of ROIs, cumulative time spent in ROIs, if animals are directing towards ROIs
    etc.

    :param str config_path: Path to SimBA project config file in Configparser format
    :param str video_name: Name of video to create feature visualizations for.
    :param dict style_attr: User-defined styles (sizes, colors etc.)
    :param int cores: Number of cores to use.

    .. note:
       `Tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/ROI_tutorial.md#part-5-visualizing-roi-features>`__.

    Examples
    ----------
    >>> style_attr = {'ROI_centers': True, 'ROI_ear_tags': True, 'Directionality': True, 'Directionality_style': 'Funnel', 'Border_color': (0, 128, 0), 'Pose_estimation': True}
    >>> _ = ROIfeatureVisualizerMultiprocess(config_path='test_/project_folder/project_config.ini', video_name='Together_1.avi', style_attr=style_attr, core_cnt=3).run()
    """

    def __init__(
        self, config_path: str, video_name: str, core_cnt: int, style_attr: dict
    ):
        super().__init__(config_path=config_path)
        if platform.system() == "Darwin":
            multiprocessing.set_start_method("spawn", force=True)

        _, self.video_name, _ = get_fn_ext(video_name)
        self.core_cnt, self.style_attr = core_cnt, style_attr
        self.save_path = os.path.join(
            self.roi_features_save_dir, self.video_name + ".mp4"
        )
        if not os.path.exists(self.roi_features_save_dir):
            os.makedirs(self.roi_features_save_dir)
        self.save_temp_dir = os.path.join(self.roi_features_save_dir, "temp")
        if os.path.exists(self.save_temp_dir):
            remove_a_folder(folder_dir=self.save_temp_dir)
        os.makedirs(self.save_temp_dir)

        self.logs_path = os.path.join(self.project_path, "logs")
        self.vid_info_df = read_video_info_csv(
            os.path.join(self.project_path, Paths.VIDEO_INFO.value)
        )
        self.roi_feature_creator = ROIFeatureCreator(config_path=config_path)
        self.file_in_path = os.path.join(
            self.outlier_corrected_dir, self.video_name + "." + self.file_type
        )
        self.video_path = os.path.join(self.project_path, "videos", video_name)
        check_file_exist_and_readable(file_path=self.file_in_path)
        self.roi_feature_creator.features_files = [self.file_in_path]
        self.roi_feature_creator.files_found = [self.file_in_path]
        self.roi_feature_creator.run()
        self.video_meta_data = get_video_meta_data(self.video_path)
        self.scalers = {}
        self.space_scale, self.radius_scale, self.res_scale, self.font_scale = (
            25,
            10,
            1500,
            0.8,
        )
        self.max_dim = max(
            self.video_meta_data["width"], self.video_meta_data["height"]
        )
        self.scalers["circle_size"] = int(
            self.radius_scale / (self.res_scale / self.max_dim)
        )
        self.scalers["font_size"] = float(
            self.font_scale / (self.res_scale / self.max_dim)
        )
        self.scalers["spacing_size"] = int(
            self.space_scale / (self.res_scale / self.max_dim)
        )
        self.data_df = read_df(self.file_in_path, self.file_type)
        self.bp_names = self.roi_feature_creator.roi_analyzer.bp_dict
        self.video_recs = self.roi_feature_creator.roi_analyzer.video_recs
        self.video_circs = self.roi_feature_creator.roi_analyzer.video_circs
        self.video_polys = self.roi_feature_creator.roi_analyzer.video_polys
        self.shape_dicts = {}
        for df in [self.video_recs, self.video_circs, self.video_polys]:
            d = df.set_index("Name").to_dict(orient="index")
            self.shape_dicts = {**self.shape_dicts, **d}
        self.video_shapes = list(
            itertools.chain(
                self.video_recs["Name"].unique(),
                self.video_circs["Name"].unique(),
                self.video_polys["Name"].unique(),
            )
        )
        self.roi_directing_viable = self.roi_feature_creator.roi_directing_viable
        if self.roi_directing_viable:
            self.directing_data = self.roi_feature_creator.directing_analyzer.results_df
            self.directing_data = self.directing_data[
                self.directing_data["Video"] == self.video_name
            ]
        else:
            self.directing_data = None
        self.roi_feature_creator.out_df.fillna(0, inplace=True)
        self.timer = SimbaTimer()
        self.timer.start_timer()

    def __calc_text_locs(self):
        add_spacer = 2
        self.loc_dict = {}
        self.cap = cv2.VideoCapture(self.video_path)
        self.cap.set(0, 1)
        ret, img = self.cap.read()
        self.img_w_border = cv2.copyMakeBorder(
            img,
            0,
            0,
            0,
            int(self.video_meta_data["width"]),
            borderType=cv2.BORDER_CONSTANT,
            value=[0, 0, 0],
        )
        self.img_w_border_h, self.img_w_border_w = (
            self.img_w_border.shape[0],
            self.img_w_border.shape[1],
        )

        for animal_cnt, animal_name in enumerate(self.multi_animal_id_list):
            self.loc_dict[animal_name] = {}
            for shape in self.video_shapes:
                self.loc_dict[animal_name][shape] = {}
                self.loc_dict[animal_name][shape]["in_zone_text"] = "{} {} {}".format(
                    shape, animal_name, "in zone"
                )
                self.loc_dict[animal_name][shape]["distance_text"] = "{} {} {}".format(
                    shape, animal_name, "distance"
                )
                self.loc_dict[animal_name][shape]["in_zone_text_loc"] = (
                    (self.video_meta_data["width"] + 5),
                    (
                        self.video_meta_data["height"]
                        - (self.video_meta_data["height"] + 10)
                        + self.scalers["spacing_size"] * add_spacer
                    ),
                )
                self.loc_dict[animal_name][shape]["in_zone_data_loc"] = (
                    int(self.img_w_border_w - (self.img_w_border_w / 8)),
                    (
                        self.video_meta_data["height"]
                        - (self.video_meta_data["height"] + 10)
                        + self.scalers["spacing_size"] * add_spacer
                    ),
                )
                add_spacer += 1
                self.loc_dict[animal_name][shape]["distance_text_loc"] = (
                    (self.video_meta_data["width"] + 5),
                    (
                        self.video_meta_data["height"]
                        - (self.video_meta_data["height"] + 10)
                        + self.scalers["spacing_size"] * add_spacer
                    ),
                )
                self.loc_dict[animal_name][shape]["distance_data_loc"] = (
                    int(self.img_w_border_w - (self.img_w_border_w / 8)),
                    (
                        self.video_meta_data["height"]
                        - (self.video_meta_data["height"] + 10)
                        + self.scalers["spacing_size"] * add_spacer
                    ),
                )
                add_spacer += 1
                if self.roi_directing_viable and self.style_attr["Directionality"]:
                    self.loc_dict[animal_name][shape]["directing_text"] = (
                        "{} {} {}".format(shape, animal_name, "facing")
                    )
                    self.loc_dict[animal_name][shape]["directing_text_loc"] = (
                        (self.video_meta_data["width"] + 5),
                        (
                            self.video_meta_data["height"]
                            - (self.video_meta_data["height"] + 10)
                            + self.scalers["spacing_size"] * add_spacer
                        ),
                    )
                    self.loc_dict[animal_name][shape]["directing_data_loc"] = (
                        int(self.img_w_border_w - (self.img_w_border_w / 8)),
                        (
                            self.video_meta_data["height"]
                            - (self.video_meta_data["height"] + 10)
                            + self.scalers["spacing_size"] * add_spacer
                        ),
                    )
                    add_spacer += 1

    def run(self):
        """
        Creates and saves visualizations of ROI-based features. Results are stored in the ``project_folder/frames/
        output/ROI_features`` directory  of the SimBA project.

        Returns
        ----------
        None
        """

        self.__calc_text_locs()
        data_arr, frm_per_core = self.split_and_group_df(
            df=self.roi_feature_creator.out_df,
            splits=self.core_cnt,
            include_split_order=True,
        )
        print(
            "Creating ROI feature images, multiprocessing (determined chunksize: {}, cores: {})...".format(
                str(self.multiprocess_chunksize), str(self.core_cnt)
            )
        )
        with multiprocessing.Pool(
            self.core_cnt, maxtasksperchild=self.maxtasksperchild
        ) as pool:
            constants = functools.partial(
                self.roi_feature_visualizer_mp,
                text_locations=self.loc_dict,
                scalers=self.scalers,
                video_meta_data=self.video_meta_data,
                shape_info=self.shape_dicts,
                style_attr=self.style_attr,
                save_temp_dir=self.save_temp_dir,
                directing_data=self.directing_data,
                video_path=self.video_path,
                directing_viable=self.roi_directing_viable,
                animal_names=self.multi_animal_id_list,
                tracked_bps=self.bp_names,
                animal_bps=self.animal_bp_dict,
            )
            for cnt, result in enumerate(
                pool.imap(constants, data_arr, chunksize=self.multiprocess_chunksize)
            ):
                print(
                    "Image {}/{}, Video {}...".format(
                        str(int(frm_per_core * (result + 1))),
                        str(len(self.data_df)),
                        self.video_name,
                    )
                )

            print("Joining {} multiprocessed video...".format(self.video_name))
            concatenate_videos_in_folder(
                in_folder=self.save_temp_dir,
                save_path=self.save_path,
                video_format="mp4",
            )
            self.timer.stop_timer()
            pool.terminate()
            pool.join()
            stdout_success(
                msg=f"Video {self.video_name} complete. Video saved in project_folder/frames/output/ROI_features.",
                elapsed_time=self.timer.elapsed_time_str,
            )


# style_attr = {'ROI_centers': True,
#               'ROI_ear_tags': True,
#               'Directionality': True,
#               'Directionality_style': 'Funnel',
#               'Border_color': (0, 128, 0),
#               'Pose_estimation': True}
# roi_feature_visualizer = ROIfeatureVisualizerMultiprocess(config_path='/Users/simon/Desktop/envs/simba_dev/tests/test_data/mouse_open_field/project_folder/project_config.ini',
#                                                           video_name='Video1.mp4',
#                                                           style_attr=style_attr,
#                                                           core_cnt=3)
# roi_feature_visualizer.create_visualization()
#
# style_attr = {'ROI_centers': True, 'ROI_ear_tags': True, 'Directionality': True, 'Directionality_style': 'Funnel', 'Border_color': (0, 128, 0), 'Pose_estimation': True}
# test = ROIfeatureVisualizerMultiprocess(config_path='/Users/simon/Desktop/envs/simba_dev/tests/test_data/mouse_open_field/project_folder/project_config.ini', video_name='Video1.mp4', style_attr=style_attr, core_cnt=5)
# test.create_visualization()
